/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "application.h"


/* hash table of the status code
   properties (message, bg etc.) */
static GHashTable *uc_application_status_code_properties_hash;

static void uc_application_menu_init (void);
static void uc_application_free_projectstate (void);
static void uc_application_make_paths (void);
static void uc_application_clean_data (void);
static void uc_application_settings_init (void);
static void uc_application_status_code_properties_free_callback (gpointer key,
								 gpointer
								 value,
								 gpointer
								 user_data);


/*
 * remove all application paths
 */
void
uc_application_remove_paths (void)
{
  DIR *dir = NULL;
  struct dirent *file = NULL;
  gchar *path = NULL;
  gchar *final_path = NULL;

  if (!
      (path =
       UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH)))
    path = g_strdup (UC_CONFIG_PROPERTIES_GENERAL_DEFAULT_WORKING_PATH);

  final_path = g_strconcat (path, "/", PACKAGE, NULL);

  if ((dir = opendir (final_path)))
    {
      while ((file = readdir (dir)))
	{
	  gchar *file_path =
	    g_strconcat (final_path, "/", file->d_name, NULL);
	  unlink (file_path);
	  g_free (file_path);
	}
      closedir (dir);
    }

  g_free (path);
  g_free (final_path);
}


/*
 * show view source dialog
 */
void
uc_application_view_source_dialog_show (void)
{
  gint id;

  if ((id = uc_check_treeview_get_selected_row_id ()) >= 0)
    {
      gchar *src = NULL;

      if ((src = uc_cache_get_source (id)))
	{
	  gchar *tmp = NULL;
	  GtkTextView *text_view_source = NULL;
	  GtkTextBuffer *buffer = NULL;

	  glade_view_source_dialog =
	    glade_xml_new (GLADE_XML_FILENAME, "view_source_dialog", NULL);
	  glade_xml_signal_autoconnect (glade_view_source_dialog);
	  view_source_dialog =
	    glade_xml_get_widget (glade_view_source_dialog,
				  "view_source_dialog");
	  text_view_source =
	    (GtkTextView *) glade_xml_get_widget (glade_view_source_dialog,
						  "textViewSource");

	  buffer =
	    gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view_source));

	  tmp = (g_utf8_validate (src, strlen (src), NULL)) ?
	    g_strdup (src) :
	    g_convert (src, strlen (src),
		       "UTF-8", "LATIN1", NULL, NULL, NULL);

	  gtk_text_buffer_set_text (buffer, tmp, -1);

	  g_free (tmp);
	  g_free (src);

	  gtk_widget_show (view_source_dialog);
	}
      else
	uc_application_dialog_show (_("The corresponding "
				      "file is empty or does not exist "
				      "in the cache directory!"),
				    GTK_MESSAGE_ERROR);
    }
  else
    uc_application_dialog_show (_("Please, select a item first."),
				GTK_MESSAGE_WARNING);
}


/*
 * free project state structure
 */
static void
uc_application_free_projectstate (void)
{
  g_free (projectState.url);
  g_free (projectState.working_path);
  g_free (projectState.proxy_config);
  g_free (projectState.proxy_host);
}


/*
 * refresh the application display
 */
void
uc_application_update_ui (void)
{
  while (gtk_events_pending ())
    gtk_main_iteration ();
}


/*
 *  free the code properties hash table
 */
void
uc_application_status_code_properties_free (void)
{
  if (!uc_application_status_code_properties_hash)
    return;

  g_hash_table_foreach (uc_application_status_code_properties_hash,
			uc_application_status_code_properties_free_callback,
			NULL);
  g_hash_table_destroy (uc_application_status_code_properties_hash);
}


/*
 *  callback for the uc_application_status_code_properties_free () function
 */
static void
uc_application_status_code_properties_free_callback (gpointer key,
						     gpointer value,
						     gpointer user_data)
{
  UCStatusCode *sc = (UCStatusCode *) value;

  g_free (sc->message);
  g_free (sc->description);
}

/*
 * -- FIXME --
 * this function load the properties to associate
 * with a server status code.
 * we can not use a external file (?) because of gettext.
 */
void
uc_application_status_code_properties_init (void)
{
  UCStatusCode *sc;
  GdkColor color;

  uc_application_status_code_properties_hash =
    g_hash_table_new (g_str_hash, g_str_equal);

  /* IGNORED ITEM */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (IGNORED_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Item ignored."));
  sc->description =
    g_strdup (_("You chose to ignore this item while gURLChecker "
		"was checking it."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_IGNORED_LINK, (gpointer) sc);

  /* UNKNOWN */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (GOOD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("No valid information."));
  sc->description =
    g_strdup (_("gURLChecker can't interpret the status code sent "
		"by the remote host.\nPlease, click on the "
		"\"Detailed display\" button, and report the value of the "
		"\"Status\" field to the author of gURLChecker."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_UNKNOWN, (gpointer) sc);

  /* 200 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (GOOD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Page available"));
  sc->description =
    g_strdup (_("No problem with this page.\nServer is up and document "
		"can be loaded without any problem."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_PAGE_OK, (gpointer) sc);

  /* 302 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (GOOD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Page available (with redirection)"));
  sc->description =
    g_strdup (_("No problem with this page.\nServer is up and document "
		"can be loaded without any problem."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_REDIRECTED, (gpointer) sc);

/* 403 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (BAD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Access forbidden."));
  sc->description =
    g_strdup (_("The remote host has refused to serve us the page."
		"\n\nTry to check it later."));
  g_hash_table_insert (uc_application_status_code_properties_hash, "403",
		       (gpointer) sc);

  /* 204 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (NO_DATA_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Page available, but with no content."));
  sc->description =
    g_strdup (_("No problem with this page, but it have no content "
		"at all...\n\nTry to check it later."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_NO_DATA, (gpointer) sc);

  /* 404 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (BAD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Page not found."));
  sc->description =
    g_strdup (_("The page does not longer exist on the server."));
  g_hash_table_insert (uc_application_status_code_properties_hash, "404",
		       (gpointer) sc);

  /* 500 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (BAD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Internal server error."));
  sc->description =
    g_strdup (_("The page contain bad code, or the server is "
		"misconfigured.\n\nTry to check it later."));
  g_hash_table_insert (uc_application_status_code_properties_hash, "500",
		       (gpointer) sc);

  /* 503 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (BAD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Service unavailable."));
  sc->description =
    g_strdup (_("The server is currently unable to handle the request "
		"due to a temporary overloading or maintenance "
		"of the server. It is also possible that you are using "
		"a proxy that can not join the given host.\n\nTry "
		"to check it later, or change your proxy settings."));
  g_hash_table_insert (uc_application_status_code_properties_hash, "503",
		       (gpointer) sc);

  /* 405 */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (BAD_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Method not allowed."));
  sc->description =
    g_strdup (_("The remote host has refused to serve us the page."
		"\n\nTry to check it later."));
  g_hash_table_insert (uc_application_status_code_properties_hash, "405",
		       (gpointer) sc);

/* TIMEOUT */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (TIMEOUT_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Receiving timout requesting the page."));
  sc->description =
    g_strdup (_("Perhaps you would change the timeout in gURLChecker "
		"preferences, or perhaps the remote host is "
		"too busy for the moment to responde to "
		"gURLChecker request.\n\nTry to check it later."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_TIMEOUT, (gpointer) sc);

  /* HOST UNREACHABLE */
  sc = g_malloc (sizeof (UCStatusCode));
  gdk_color_parse (HOST_UNREACHABLE_LINK_BGCOLOR, &color);
  sc->background = color;
  sc->message = g_strdup (_("Host unreachable."));
  sc->description =
    g_strdup (_("The remote host does not respond at all...\nPerhaps "
		"you would change the timeout in gURLChecker "
		"preferences, or perhaps it is really down."
		"\n\nTry to check it later."));
  g_hash_table_insert (uc_application_status_code_properties_hash,
		       STATUS_CODE_HOST_IS_UNREACHABLE, (gpointer) sc);
}


/*
 * active or deactive the manual proxy configuration
 * zone on the settings dialog box
 */
void
uc_application_proxy_frame_control_sensitive (void)
{
  GConfClient *client = NULL;
  GtkWidget *widget_host = NULL;
  GtkWidget *widget_port = NULL;
  GtkWidget *widget_proxy_gnome = NULL;
  GtkWidget *widget_proxy_manual = NULL;
  GtkWidget *widget_proxy_none = NULL;

  client = gconf_client_get_default ();
  widget_host =
    glade_xml_get_widget (glade_settings_dialog, "entryProxyHost");
  widget_port = glade_xml_get_widget (glade_settings_dialog, "spinProxyPort");
  widget_proxy_gnome =
    glade_xml_get_widget (glade_settings_dialog, "radio_proxy_gnome");
  widget_proxy_manual =
    glade_xml_get_widget (glade_settings_dialog, "radio_proxy_manual");
  widget_proxy_none =
    glade_xml_get_widget (glade_settings_dialog, "radio_proxy_none");

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget_proxy_none)))
    {
      gtk_widget_set_sensitive (glade_xml_get_widget
				(glade_settings_dialog, "table_proxy_params"),
				FALSE);

      return;
    }

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget_proxy_gnome)))
    {
      if (!gconf_client_get_bool
	  (client, "/system/http_proxy/use_http_proxy", NULL))
	{
	  uc_application_dialog_show (_("You must activate your GNOME "
					"proxy propertie with the Gnome Control "
					"Center  before doing this."),
				      GTK_MESSAGE_ERROR);
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget_proxy_none),
					TRUE);
	}
      else
	{
	  gchar *host = NULL;
	  guint port = 0;

	  host =
	    gconf_client_get_string (client, "/system/http_proxy/host", NULL);
	  gtk_entry_set_text (GTK_ENTRY (widget_host), host);
	  g_free (host);

	  port =
	    gconf_client_get_int (client, "/system/http_proxy/port", NULL);
	  gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget_port),
				     (gfloat) port);
	  gtk_widget_set_sensitive (glade_xml_get_widget
				    (glade_settings_dialog,
				     "table_proxy_params"), FALSE);
	}

      return;
    }

  gtk_widget_set_sensitive (glade_xml_get_widget
			    (glade_settings_dialog, "table_proxy_params"),
			    TRUE);
}


/*
 *  show the settings dialog box
 */
void
uc_application_settings_dialog_show (void)
{
  GtkWidget *widget = NULL;
  GtkWidget *widget_bis = NULL;
  gchar **items = NULL;
  gchar **selected_items = NULL;
  gchar *str = NULL;
  GtkWidget *frame = NULL;
  guint items_size = 0;
  guint value = 0;
  guint selected_items_size = 0;

  glade_settings_dialog =
    glade_xml_new (GLADE_XML_FILENAME, "settings_dialog", NULL);
  glade_xml_signal_autoconnect (glade_settings_dialog);
  settings_dialog =
    glade_xml_get_widget (glade_settings_dialog, "settings_dialog");

  /* GENERAL */

  /* printing format */
  widget =
    glade_xml_get_widget (glade_settings_dialog, "radioDetailedPrinting");
  widget_bis =
    glade_xml_get_widget (glade_settings_dialog, "radioBasicPrinting");
  str = UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_DISPLAY_FORMAT);

  if (str)
    {
      if (!strcmp ("detailed", str))
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget_bis), TRUE);
    }
  g_free (str);

  /* network timeout */
  widget = glade_xml_get_widget (glade_settings_dialog, "spinTimeout");
  value = UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_TIMEOUT);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gfloat) value);

  /* proxy */
  widget = glade_xml_get_widget (glade_settings_dialog, "entryProxyHost");
  if ((str =
       UC_CONFIG_GET_STRING
       (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST)))
    {
      gtk_entry_set_text (GTK_ENTRY (widget), str);
      g_free (str);
    }

  widget_bis = glade_xml_get_widget (glade_settings_dialog, "spinProxyPort");
  value =
    UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget_bis), (gfloat) value);

  str =
    UC_CONFIG_GET_STRING
    (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG);
  if (str)
    {
      if (!strcmp ("none", str))
	{
	  widget =
	    glade_xml_get_widget (glade_settings_dialog, "radio_proxy_none");
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
	  gtk_widget_set_sensitive (glade_xml_get_widget
				    (glade_settings_dialog,
				     "table_proxy_params"), FALSE);
	}
      else if (!strcmp ("gnome", str))
	{
	  widget =
	    glade_xml_get_widget (glade_settings_dialog, "radio_proxy_gnome");
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
	  gtk_widget_set_sensitive (glade_xml_get_widget
				    (glade_settings_dialog,
				     "table_proxy_params"), FALSE);
	}
      else
	{
	  widget =
	    glade_xml_get_widget (glade_settings_dialog,
				  "radio_proxy_manual");
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
	}
    }
  g_free (str);

  /* working path */
  widget = glade_xml_get_widget (glade_settings_dialog, "workingPath");
  gnome_file_entry_set_title (GNOME_FILE_ENTRY (widget),
			      _("Choose a directorie"));
  widget_bis =
    glade_xml_get_widget (glade_settings_dialog, "workingPathEntry");
  str = UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH);
  if (str)
    {
      gtk_entry_set_text (GTK_ENTRY (widget_bis), str);
      gnome_file_entry_set_default_path (GNOME_FILE_ENTRY (widget), str);
    }
  else
    {
      gtk_entry_set_text (GTK_ENTRY (widget_bis),
			  UC_CONFIG_PROPERTIES_GENERAL_DEFAULT_WORKING_PATH);
      gnome_file_entry_set_default_path (GNOME_FILE_ENTRY (widget),
					 UC_CONFIG_PROPERTIES_GENERAL_DEFAULT_WORKING_PATH);
    }
  g_free (str);

  /* URL */

  /* networking check level */
  widget = glade_xml_get_widget (glade_settings_dialog, "spinURLCheckLevel");
  value = UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_CHECK_LEVEL);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gfloat) value);

  /* networking depth level */
  widget = glade_xml_get_widget (glade_settings_dialog, "spinURLDepthLevel");
  value = UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_DEPTH_LEVEL);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gfloat) value);

  /* documents to reject */
  frame = glade_xml_get_widget (glade_settings_dialog, "frameURLDocuments");
  items = UC_CONFIG_PROPERTIES_GET_DEFAULT_CONTENT_TYPES_DOCUMENTS;
  items_size = UC_CONFIG_PROPERTIES_DEFAULT_CONTENT_TYPES_DOCUMENTS_LEN;
  UC_CONFIG_GET_VECTOR (UC_CONFIG_PROPERTIES_URL_CONTENT_TYPES_DOCUMENTS,
			&selected_items, &selected_items_size);
  uc_check_buttons_display ((GtkWidget *) frame, items, items_size,
			    selected_items, selected_items_size);
  g_strfreev (items);

  /* images to reject */
  frame = glade_xml_get_widget (glade_settings_dialog, "frameURLImages");
  items = UC_CONFIG_PROPERTIES_GET_DEFAULT_CONTENT_TYPES_IMAGES;
  items_size = UC_CONFIG_PROPERTIES_DEFAULT_CONTENT_TYPES_IMAGES_LEN;
  UC_CONFIG_GET_VECTOR (UC_CONFIG_PROPERTIES_URL_CONTENT_TYPES_IMAGES,
			&selected_items, &selected_items_size);
  uc_check_buttons_display (frame, items, items_size, selected_items,
			    selected_items_size);
  g_strfreev (items);

  gtk_widget_show (settings_dialog);
}


/*
 *  retrieving user input data settings
 */
gboolean
uc_application_settings_get_data (void)
{
  GtkWidget *widget = NULL;
  gchar *str = NULL;
  gchar *host = NULL;
  guint port = 0;
  gint sock = 0;

  /* GENERAL */

  /* printing format */
  widget =
    glade_xml_get_widget (glade_settings_dialog, "radioDetailedPrinting");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
    UC_CONFIG_SET_STRING (UC_CONFIG_PROPERTIES_GENERAL_DISPLAY_FORMAT,
			  "detailed");
  else
    UC_CONFIG_SET_STRING (UC_CONFIG_PROPERTIES_GENERAL_DISPLAY_FORMAT,
			  "basic");

  /* network timeout */
  widget = glade_xml_get_widget (glade_settings_dialog, "spinTimeout");
  UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_TIMEOUT, (gint)
		     gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget)));

  /* proxy */
  widget = glade_xml_get_widget (glade_settings_dialog, "radio_proxy_none");
  if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
    {
      widget =
	glade_xml_get_widget (glade_settings_dialog, "radio_proxy_gnome");
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
	UC_CONFIG_SET_STRING
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG, "gnome");
      else
	UC_CONFIG_SET_STRING
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG, "manual");

      widget = glade_xml_get_widget (glade_settings_dialog, "entryProxyHost");
      host = (gchar *) gtk_entry_get_text (GTK_ENTRY (widget));
      widget = glade_xml_get_widget (glade_settings_dialog, "spinProxyPort");
      port = (gint) gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));

      if ((!strlen (host)) ||
	  ((sock =
	    uc_utils_socket_open (host, port,
				  UC_SOCKET_OPEN_FORCE_ARGS_USE)) <= 0))
	{
	  uc_application_dialog_show (_("The proxy host:port you entered "
					"does not respond. Please, change "
					"those parameters."),
				      GTK_MESSAGE_ERROR);
	  return FALSE;
	}
      close (sock);
      UC_CONFIG_SET_STRING
	(UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST, host);
      UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT,
			 port);
    }
  else
    UC_CONFIG_SET_STRING
      (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG, "none");

  /* working path */
  widget = glade_xml_get_widget (glade_settings_dialog, "workingPath");
  if (!
      (str =
       gnome_file_entry_get_full_path (GNOME_FILE_ENTRY (widget), FALSE)))
    {
      uc_application_dialog_show (_("Please choose a working path for "
				    "gURLChecker."), GTK_MESSAGE_ERROR);
      return FALSE;
    }

  if (!uc_utils_working_path_is_valid (str))
    {
      uc_application_dialog_show (_("The path you entered is not a "
				    "directory or your don't have appropriates "
				    "rights on it. Please choose a valid "
				    "working path for gURLChecker."),
				  GTK_MESSAGE_ERROR);
      return FALSE;
    }

  UC_CONFIG_SET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH, str);

  /* URL */

  /* networking check level */
  widget = glade_xml_get_widget (glade_settings_dialog, "spinURLCheckLevel");
  UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_CHECK_LEVEL,
		     (gint)
		     gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget)));

  /* networking depth level */
  widget = glade_xml_get_widget (glade_settings_dialog, "spinURLDepthLevel");
  UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_DEPTH_LEVEL,
		     (gint)
		     gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget)));

  /* documents to reject */
  widget = glade_xml_get_widget (glade_settings_dialog, "frameURLDocuments");
  uc_check_buttons_save_setting (widget,
				 UC_CONFIG_PROPERTIES_URL_CONTENT_TYPES_DOCUMENTS);

  /* images to reject */
  widget = glade_xml_get_widget (glade_settings_dialog, "frameURLImages");
  uc_check_buttons_save_setting (widget,
				 UC_CONFIG_PROPERTIES_URL_CONTENT_TYPES_IMAGES);
  gnome_config_sync ();

  return TRUE;
}


/*
 * show the progess dialog
 */
void
uc_application_progress_dialog_show (void)
{
  glade_progress_dialog =
    glade_xml_new (GLADE_XML_FILENAME, "progress_dialog", NULL);
  glade_xml_signal_autoconnect (glade_progress_dialog);
  progress_dialog =
    glade_xml_get_widget (glade_progress_dialog, "progress_dialog");

  gtk_widget_show (progress_dialog);
}


/*
 *  show the about box
 */
void
uc_application_about_dialog_show (void)
{
  GtkWidget *about_dialog = NULL;
  const gchar *authors[] = {
    "Emmanuel Saracco <esaracco@noos.fr>",
    NULL
  };

  about_dialog =
    gnome_about_new (PACKAGE,
		     VERSION,
		     "(C) 2002,2003 Emmanuel Saracco <esaracco@noos.fr>",
		     _("gURLChecker can check links on a single web page "
		       "or on a whole web site in order to "
		       "determine validity of each page."),
		     authors, NULL, NULL, NULL);

  gtk_window_set_destroy_with_parent (GTK_WINDOW (about_dialog), TRUE);
  gtk_widget_show (about_dialog);
}


/*
 *  print a alert box
 */
void
uc_application_dialog_show (const gchar * message,
			    const GtkMessageType msg_type)
{
  GtkWidget *dialog = gtk_message_dialog_new (NULL,
					      GTK_DIALOG_MODAL |
					      GTK_DIALOG_DESTROY_WITH_PARENT,
					      msg_type,
					      GTK_BUTTONS_OK,
					      message);

  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
  gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);
}



/*
 *  show the new project dialog box
 */
void
uc_application_new_dialog_show (void)
{

  glade_new_project_dialog =
    glade_xml_new (GLADE_XML_FILENAME, "new_project_dialog", NULL);
  glade_xml_signal_autoconnect (glade_new_project_dialog);

  new_project_dialog =
    glade_xml_get_widget (glade_new_project_dialog, "new_project_dialog");

  gtk_widget_show (new_project_dialog);
}


/*
 *  retrieving user data input preferences
 */
void
uc_application_get_data (void)
{
  GtkWidget *entryProjectURL =
    glade_xml_get_widget (glade_new_project_dialog, "entryProjectURL");
  gchar *tmp = NULL;

  tmp = (gchar *) gtk_entry_get_text (GTK_ENTRY (entryProjectURL));

  if (tmp)
    {
      g_strstrip (tmp);
      if (strlen (tmp))
	{
	  gchar *tmp_url = uc_url_add_protocol (PROTOCOL_HTTP, tmp);

	  if (uc_url_is_valid (tmp_url))
	    {
	      g_free (projectState.url);
	      projectState.url = tmp_url;

	      uc_application_properties_init ();
	      gtk_widget_destroy (new_project_dialog);
	      uc_application_update_ui ();
	      uc_check_begin ();

	      return;
	    }
	}
    }

  uc_application_dialog_show (_("Please, check that you "
				"entered the URL correctly and try again."),
			      GTK_MESSAGE_ERROR);
}


/*
 * draw the two principals main frames
 * for displaying urls list and headers
 */
void
uc_application_draw_main_frames ()
{
  GtkWidget *main_frames = NULL;
  GtkWidget *right_frames = NULL;
  GtkWidget *urls_list = NULL;
  GtkWidget *scrolled_window = NULL;
  GtkWidget *label = NULL;
  guint w = 0;
  guint h = 0;

  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window, "main_frames",
			main_frames);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window, "right_frames",
			right_frames);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window, "scrolled_window",
			scrolled_window);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window,
			"label_informations_message", label);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window,
			"labelReportCheckedLinks", label);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window, "labelReportBadLinks",
			label);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window,
			"labelReportGoodLinks", label);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window,
			"labelReportTimedOutLinks", label);
  UC_GET_CUSTOM_WIDGET (glade_main_window, main_window,
			"labelReportElapsedTime", label);

  if ((urls_list = UC_GET_CUSTOM_WIDGET_REAL (main_window, "urls_list")))
    gtk_widget_destroy (urls_list);

  urls_list = gtk_tree_view_new ();
  gtk_widget_show (urls_list);
  gtk_container_add (GTK_CONTAINER (scrolled_window), urls_list);
  UC_SET_CUSTOM_WIDGET (main_window, "urls_list", urls_list);

  g_signal_connect (G_OBJECT (urls_list), "cursor_changed",
		    G_CALLBACK (on_urlsList_cursorChanged), NULL);

  gtk_window_get_size (GTK_WINDOW (main_window), &w, &h);
  gtk_paned_set_position (GTK_PANED (main_frames), (guint) (h / 2));
  gtk_paned_set_position (GTK_PANED (right_frames),
			  (guint) ((w / 2) + (w / 30)));

  gtk_widget_show (right_frames);
  gtk_widget_show (main_frames);
}


/*
 * initialize some global project properties
 */
void
uc_application_properties_init (void)
{
  projectState.display_format =
    (!strcmp
     (UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_DISPLAY_FORMAT),
      "basic")) ? SIMPLE_PRINTING : DETAILED_PRINTING;
  projectState.check_level =
    UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_CHECK_LEVEL);
  projectState.depth_level =
    UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_DEPTH_LEVEL);
  g_free (projectState.working_path);
  projectState.working_path =
    UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH);

  g_free (projectState.proxy_config);
  projectState.proxy_config =
    UC_CONFIG_GET_STRING
    (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG);
  g_free (projectState.proxy_host);
  projectState.proxy_host =
    UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST);
  projectState.proxy_port =
    UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT);
}


/*
 * return properties for a given status code
 */
UCStatusCode *
uc_application_get_status_code_properties (const gchar * status_code)
{
  return g_hash_table_lookup (uc_application_status_code_properties_hash,
			      status_code);
}


/*
 *  do a clean application exit
 */
void
uc_application_quit (void)
{
  uc_application_clean_data ();
  gtk_exit (0);
}


/*
 *  general clean up routine call at application
 *  exit
 */
static void
uc_application_clean_data (void)
{
  uc_application_remove_paths ();
  uc_application_free_projectstate ();
  uc_check_free_link_properties_list (uc_link_properties_list);
  uc_check_free_tmp_list ();
  uc_application_status_code_properties_free ();

  gnome_config_sync ();
}

/*
 *  gURLChecker initialization
 */
void
uc_application_init (void)
{
  uc_application_menu_init ();
  uc_application_remove_paths ();
  uc_application_make_paths ();
  uc_application_status_code_properties_init ();
  uc_application_settings_init ();
  uc_application_properties_init ();
  uc_check_reset ();
}


/*
 *  create the temporaries gURLChecker directories
 */
static void
uc_application_make_paths (void)
{
  gchar *path = NULL;
  gchar *final_path = NULL;

  if (!
      (path =
       UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH)))
    path = g_strdup (UC_CONFIG_PROPERTIES_GENERAL_DEFAULT_WORKING_PATH);

  final_path = g_strconcat (path, "/", PACKAGE, NULL);
  mkdir (final_path, S_IRWXU);

  g_free (final_path);
  g_free (path);
}


/*
 * set the defaults properties the first time
 * gURLChecker is launched
 */
static void
uc_application_settings_init (void)
{
  gchar *str = NULL;
  GConfClient *client = NULL;

  client = gconf_client_get_default ();

  /* display format */
  if (!
      (str =
       UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_DISPLAY_FORMAT)))
    UC_CONFIG_SET_STRING (UC_CONFIG_PROPERTIES_GENERAL_DISPLAY_FORMAT,
			  "basic");
  else
    g_free (str);

  /* networking timeout */
  if (!UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_TIMEOUT))
    UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_TIMEOUT,
		       CHECK_TIMEOUT_DEFAULT);

  /* check level */
  if (!UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_CHECK_LEVEL))
    UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_CHECK_LEVEL,
		       CHECK_LEVEL_DEFAULT);

  /* working path */
  if (!
      (str =
       UC_CONFIG_GET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH)))
    UC_CONFIG_SET_STRING (UC_CONFIG_PROPERTIES_GENERAL_WORKING_PATH,
			  UC_CONFIG_PROPERTIES_GENERAL_DEFAULT_WORKING_PATH);
  else
    g_free (str);

  /* proxy */
  if (gconf_client_get_bool
      (client, "/system/http_proxy/use_http_proxy", NULL))
    {
      gchar *host = NULL;
      guint port = 0;

      if (!
	  (str =
	   UC_CONFIG_GET_STRING
	   (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG)))
	UC_CONFIG_SET_STRING
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG, "gnome");
      else
	g_free (str);

      if (!
	  (str =
	   UC_CONFIG_GET_STRING
	   (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST)))
	{
	  host =
	    gconf_client_get_string (client, "/system/http_proxy/host", NULL);
	  UC_CONFIG_SET_STRING
	    (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST, host);
	  g_free (host);
	}
      else
	g_free (str);

      if (!UC_CONFIG_GET_INT
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT))
	{
	  port =
	    gconf_client_get_int (client, "/system/http_proxy/port", NULL);
	  UC_CONFIG_SET_INT
	    (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT, port);
	}
    }
  else
    {
      if (!
	  (str =
	   UC_CONFIG_GET_STRING
	   (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG)))
	UC_CONFIG_SET_STRING
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_CONFIG, "none");
      else
	g_free (str);

      if (!
	  (str =
	   UC_CONFIG_GET_STRING
	   (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST)))
	UC_CONFIG_SET_STRING
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_HOST, "");
      else
	g_free (str);

      if (!UC_CONFIG_GET_INT
	  (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT))
	UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_GENERAL_NETWORKING_PROXY_PORT,
			   PROXY_PORT_DEFAULT);
    }

  /* depth level */
  if (!UC_CONFIG_GET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_DEPTH_LEVEL))
    UC_CONFIG_SET_INT (UC_CONFIG_PROPERTIES_URL_NETWORKING_DEPTH_LEVEL,
		       DEPTH_LEVEL_DEFAULT);

  gnome_config_sync ();
}


/*
 * first menu items management
 */
static void
uc_application_menu_init (void)
{
  /* no treeview, then no display or edit menu options */
  UC_WIDGET_SET_SENSITIVE (glade_main_window, "display", FALSE);
  UC_WIDGET_SET_SENSITIVE (glade_main_window, "edit", FALSE);
}
