/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "url.h"


/*
 * add the given protocol to the given url
 */
gchar *
uc_url_add_protocol (const gchar * proto, const gchar * host)
{
  if ((strstr (host, "://")))
    return g_strdup (host);

  return g_strconcat (proto, "://", host, NULL);
}


/*
 * get the hostname of a given url.
 * if the url doesn't have host name, then
 * current_host is send
 */
gchar *
uc_url_get_hostname (const gchar * current_host, const gchar * url)
{
  gchar **split_proto = NULL;
  gchar **split_slash = NULL;
  gchar *tmp = NULL;

  if ((!url) || (!strlen (url)) || url[0] == '/')
    return (current_host) ? g_strdup (current_host) : NULL;

  split_proto = g_strsplit (url, "://", 2);
  if (!split_proto[1])
    {
      g_strfreev (split_proto);
      return (current_host) ? g_strdup (current_host) : NULL;
    }

  split_slash = g_strsplit (split_proto[1], "/", 2);

  if (!strchr (split_slash[0], ':'))
    tmp = g_strdup (split_slash[0]);
  else
    {
      gchar **split_port = NULL;

      split_port = g_strsplit (split_slash[0], ":", 2);
      tmp = g_strdup (split_port[0]);
      g_strfreev (split_port);
    }

  g_strfreev (split_proto);
  g_strfreev (split_slash);

  return tmp;
}


/*
 * give the ip of a given host
 */
gchar *
uc_url_get_ip (gchar * host)
{
  struct hostent *serverHostEnt = NULL;
  gchar *buffer = NULL;

  if ((glong) (inet_addr (host)) != (glong) - 1)
    buffer = g_strdup (host);
  else if ((serverHostEnt = gethostbyname (host)) != NULL)
    buffer =
      g_strdup (inet_ntoa
		(*((struct in_addr *) serverHostEnt->h_addr_list[0])));

  return buffer;
}


/*
 * parse a given url
 */
void
uc_url_parse (const gchar * current_host, const gchar * current_path,
	      const gchar * url, gchar ** host, gchar ** port, gchar ** dir,
	      gchar ** file, gchar ** args)
{
  const gchar *p = NULL;
  const gchar *host_b = NULL, *host_e = NULL;
  const gchar *path_b = NULL, *path_e = NULL;
  const gchar *port_b = NULL, *port_e = NULL;
  const gchar *last_slash = NULL;
  gchar *path = NULL;
  gboolean have_scheme = FALSE;
  gchar *str = NULL;

  if (strchr (url, '#'))
    {
      gchar **split = NULL;

      split = g_strsplit (url, "#", 2);
      str = g_strdup (split[0]);

      g_strfreev (split);
    }
  else
    str = g_strdup (url);


  if (strlen (str))
    {

      if ((p = strstr (str, "://")))
	{
	  p += 3;
	  have_scheme = TRUE;

	  host_b = p;
	  p = uc_utils_strpbrk_or_eos (p, ":/;?#");
	  host_e = p;
	  *host = uc_utils_strdup_delim (host_b, host_e);

	  if (*p == ':')
	    {
	      port_b = ++p;
	      p = uc_utils_strpbrk_or_eos (p, "/;?#");
	      port_e = p;
	      *port = uc_utils_strdup_delim (port_b, port_e);
	    }
	}
      else
	{
	  p = str;
	  *host = g_strdup (current_host);
	}

      if (strchr (p, '?'))
	{
	  gchar **split = NULL;

	  split = g_strsplit (p, "?", 2);
	  *args = g_strdup_printf ("?%s", split[1]);

	  g_strfreev (split);
	  ((gchar *) p)[strcspn (p, "?")] = '\0';
	}
      else
	*args = g_strdup ("");

      if (!port_b)
	*port = g_strdup (URL_DEFAULT_PORT);

      if (*p == '/')
	{
	  path_b = p;
	  p = uc_utils_strpbrk_or_eos (p, ";?#");
	  path_e = p;
	}
      else
	path_b = path_e = p;

      path = uc_utils_strdup_delim (path_b, path_e);

      last_slash = strrchr (path, '/');
      if (!last_slash)
	{
	  if (!have_scheme)
	    {
	      *dir = g_strdup (current_path);
	      *file = (!have_scheme) ? g_strdup (str) : g_strdup (path);
	    }
	  else
	    {
	      *dir = g_strdup ("");
	      *file = g_strdup ("");
	    }
	}
      else
	{
	  *dir = uc_utils_strdup_delim (path, last_slash);
	  *file = g_strdup (last_slash);
	}

      if (!have_scheme)
	{
	  if ((!strlen (*dir)) && (!strlen (*file)))
	    {
	      if (strchr (*host, '.'))
		{
		  *file = g_realloc (*file, strlen (*host) + 1);
		  strcat (*file, *host);
		}
	      else
		{
		  g_free (*dir);
		  *dir = g_strdup (*host);
		}

	      g_free (*host);
	      *host = g_strdup (current_host);
	    }

	  if (g_strcasecmp (current_host, *host))
	    {
	      gchar *tmp = NULL;

	      tmp = g_strconcat (*host, *dir, NULL);
	      g_free (*host);
	      g_free (*dir);

	      *host = g_strdup (current_host);
	      *dir = tmp;
	    }
	}

      if (!strlen (*dir) && (strlen (*file)) && (!strchr (*file, '/')))
	{
	  g_free (*dir);
	  if (strlen (current_path))
	    *dir = g_strdup (current_path);
	  else
	    *dir = g_strdup ("/");
	}

      if ((!strlen (*dir)) &&
	  (!strlen (*file)) && ((*host)[strlen (*host) - 1] != '/'))
	{
	  g_free (*dir);
	  *dir = g_strdup ("/");
	}

      if ((strlen (*dir)) && ((*dir)[strlen (*dir) - 1] != '/')
	  && ((*file)[0] != '/'))
	{
	  gchar *tmp = NULL;

	  tmp = g_strdup_printf ("%s/", *dir);
	  g_free (*dir);
	  *dir = tmp;
	}

      if (!strchr (*file, '.'))
	{
	  gchar *tmp = NULL;

	  tmp = g_strconcat ((*dir) ? *dir : "", *file, NULL);
	  g_free (*dir);
	  g_free (*file);

	  *file = g_strdup ("");
	  *dir = tmp;
	}
    }
  else
    {
      *host = g_strdup (current_host);
      *dir = g_strdup ("/");
      *file = g_strdup ("");
      *port = g_strdup (URL_DEFAULT_PORT);
      *args = g_strdup ("");
    }

  if (!args)
    *args = g_strdup ("");
  if (!file)
    *file = g_strdup ("");
  if (!dir)
    *dir = g_strdup ("");
  if (!host)
    *host = g_strdup ("");
  if (!port)
    *port = g_strdup ("");

  g_free (str);
  g_free (path);
}


/*
 * get the protocol of a given url
 */
gchar *
uc_url_get_protocol (const gchar * url)
{
  gchar **split = NULL;
  gchar *tmp = NULL;

  if (!strchr (url, ':'))
    return g_strup (g_strdup (PROTOCOL_HTTP));

  split = g_strsplit (url, ":", 0);
  tmp = g_strup (g_strdup (split[0]));
  g_strfreev (split);

  return tmp;
}


/*
 * check the validity of a given url
 */
gboolean
uc_url_is_valid (const gchar * url)
{
  gchar *test = NULL;
  gchar *host = NULL;

  if ((!url) || (!strlen (url))
      || (!(host = uc_url_get_hostname (NULL, url))))
    return FALSE;

  if (!(test = uc_url_get_ip (host)))
    {
      g_free (host);
      return FALSE;
    }

  g_free (host);
  g_free (test);

  return TRUE;
}
