/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "utils.h"


/*
 * true if it exist enough space on the parition
 * of the given path to write on it
 */
gboolean
uc_utils_disk_free_enough (gchar * path, guint size)
{
  return TRUE;
}


/*
 * clean a link value string
 */
gchar *
uc_utils_clean_tag_link_value (const gchar * value)
{
  gchar tmp[BUFFER_URL_LEN + 1] = { 0 };
  guint i = 0;
  guint j = 0;

  if (!value)
    return g_strdup ("");

  for (j = 0; j < strlen (value); j++)
    if ((i < BUFFER_URL_LEN)
	&& (value[j] != '\'' && value[j] != '"' && value[j] != '>'))
      tmp[i++] = value[j];

  return g_strdup (tmp);
}


/*
 * return a pointer on the next char after the searched string
 * if it found, else return the original pointer
 */
gpointer
uc_utils_search_string_next (gpointer buf, const gchar * str,
			     const gchar limit_char)
{
  gpointer sav = buf;
  gchar *buffer = (gchar *) buf;

  if ((!str) || (!strlen (str)) || (!buffer)
      || (strlen (buffer) < strlen (str)))
    return (gpointer) buf;

  while ((!uc_utils_memcasecmp (buffer, str)) && (*buffer != limit_char))
    buffer++;

  if (*buffer != limit_char)
    sav = (buffer += strlen (str));

  return sav;
}



/*
 * block signals
 */
void
uc_utils_block_signals (void)
{
  sigset_t set;

  sigemptyset (&set);
  sigaddset (&set, SIGUSR1);
  sigprocmask (SIG_BLOCK, &set, NULL);
}


/*
 * unblock signals
 */
void
uc_utils_unblock_signals (void)
{
  sigset_t set;

  sigemptyset (&set);
  sigaddset (&set, SIGUSR1);
  sigprocmask (SIG_UNBLOCK, &set, NULL);
}


/*
 * comparison between two strings, based on the length of
 * the second string
 */
gboolean
uc_utils_memcasecmp (const gchar * str1, const gchar * str2)
{
  gchar tmp[BUFFER_URL_LEN] = { 0 };
  guint len = strlen (str2);

  if ((len > BUFFER_URL_LEN) || (strlen (str1) < len))
    return FALSE;

  if (!memcmp (str1, str2, len))
    return TRUE;

  g_memmove (tmp, str1, len);
  return (!g_strcasecmp (tmp, str2));
}


/*
 * check if the given path is a valid
 * for gURLChecker work 
 */
gboolean
uc_utils_working_path_is_valid (const gchar * path)
{
  gchar test_path[BUFFER_URL_LEN + 1] = { 0 };
  gboolean ret = FALSE;
  gint fd;

  if ((snprintf (test_path, BUFFER_URL_LEN, "%s/%s/filetest", path, PACKAGE) <
       0))
    return FALSE;

  if ((fd = open (test_path, O_CREAT, S_IRUSR | S_IWUSR)) >= 0)
    {
      close (fd);
      unlink (test_path);
      ret = TRUE;
    }

  return ret;
}


/*
 * open a socket on the given host and port
 */
guint
uc_utils_socket_open (gchar * host, const guint port,
		      const gboolean force_args_use)
{
  struct sockaddr_in server_sockaddr;
  struct hostent *serverHostEnt = NULL;
  gchar buffer[BUFFER_URL_LEN + 1] = { 0 };
  gchar *host_real = NULL;
  guint port_real = 0;
  guint sock = 0;
  gulong hostAddr = 0;
  gboolean ok = FALSE;

  /* current informations for timeouts callback */
  uc_check_currentitem_set_host (host);

  sock = socket (AF_INET, SOCK_STREAM, 0);

  if (sock <= 0)
    return 0;

  if ((force_args_use == UC_SOCKET_OPEN_FORCE_ARGS_USE) ||
      (!strcmp (projectState.proxy_config, "none")))
    {
      host_real = host;
      port_real = port;
    }
  else
    {
      host_real = projectState.proxy_host;
      if (strlen (host_real))
	port_real = projectState.proxy_port;
      else
	{
	  host_real = host;
	  port_real = port;
	}
    }

  memset (&server_sockaddr, 0, sizeof (server_sockaddr));
  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons (port_real);

  uc_check_alarm_add (CHECK_TIMEOUT, &uc_check_alarm_callback);
  if ((glong) (hostAddr = inet_addr (host_real)) != (glong) - 1)
    memcpy (&server_sockaddr.sin_addr, &hostAddr, sizeof (hostAddr));
  else if ((serverHostEnt = gethostbyname (host_real)) != NULL)
    {
      inet_ntop (AF_INET,
		 (struct in_addr *) serverHostEnt->h_addr_list[0], buffer,
		 BUFFER_URL_LEN);
      inet_pton (AF_INET, buffer, &server_sockaddr.sin_addr);
    }

  if (serverHostEnt)
    ok = (!connect (sock, (struct sockaddr *) &server_sockaddr,
		    sizeof (server_sockaddr)));

  uc_check_alarm_remove ();

  return (ok) ? sock : 0;
}


/*
 * make a string with 2 given
 * pointers
 */
gchar *
uc_utils_strdup_delim (const gchar * begin, const gchar * end)
{
  gchar *res = NULL;

  res = (gchar *) g_malloc (end - begin + 1);
  memcpy (res, begin, end - begin);
  res[end - begin] = '\0';

  return res;
}


/*
 * return a pointer on a given char address
 * or on the end of the passed string
 */
gchar *
uc_utils_strpbrk_or_eos (const gchar * str, const gchar * accept)
{
  gchar *p = NULL;

  p = strpbrk (str, accept);
  if (!p)
    p = (gchar *) str + strlen (str);

  return p;
}


/*
 * return TRUE if the first argument is in
 * the second argument vector
 */
gboolean
uc_utils_array_is_in (const gchar * item, gpointer array, const guint size)
{
  gchar **items = (gchar **) array;
  guint i = 0;

  for (i = 0; i < size; i++)
    if (!strcmp (item, items[i]))
      return TRUE;

  return FALSE;
}
