/*
 * Copyright (C) 2002-2003 Stefan Holst
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: disc.c,v 1.9 2003/03/22 10:25:49 skaboy Exp $
 *
 * autoplay discs of different types
 */

/*
 * some ideas of autodetection from Olivier Dubroca, Damien Morel in
 * LOMB-Project (http://lomb.abeille.com) (c) 2002. Their code was 
 * originally written by Gerd Knorr <kraxel@goldbach.in-berlin.de> and
 * Heiko Eissfeldt <heiko@colossus.escape.de> (c) 1996,1997,1998.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/cdrom.h>

#include "oxine.h"
#include "disc.h"
#include "utils.h"

#define LOG

int autocd_mount(const char *mountpoint) {

  char command[256];
  
  snprintf(command, 256, "mount %s", mountpoint);
#ifdef LOG
    printf("autocd: mounting cdrom...\n");
#endif

  if(system(command)) {
    printf("autocd: mount failed\n");
    return 0;
  }

  return 1;
}

int autocd_umount(const char *mountpoint) {

  char command[256];
  
  snprintf(command, 256, "umount %s", mountpoint);
#ifdef LOG
    printf("autocd: unmounting cdrom...\n");
#endif

  if(system(command)) {
    printf("autocd: umount failed\n");
    return 0;
  }

  return 1;
}

int autocd_setspeed(const char *device, int speed) {

  int fd;
  int ret;

  fd = open(device, O_RDONLY | O_NONBLOCK);
  if (fd<0) {
    printf("autocd: error opening device\n");
    return 0;
  }

  ret = ioctl(fd, CDROM_SELECT_SPEED, speed);
  
  if (ret < 0) {
    printf("autocd: set speed failed\n");
    close(fd);
    return 0;
  }
  
  close(fd);

  return 1;
}

int autocd_disc_in(const char *device) {

  int fd;
  int drive_status;
  
  fd = open(device, O_RDONLY | O_NONBLOCK);
  if (fd<0) {
    printf("autocd: error opening device\n");
    return 0;
  }
  
  drive_status = ioctl (fd, CDROM_DRIVE_STATUS);

#ifdef LOG
  printf("autocd: drive reports: ");
#endif

  switch(drive_status) {
    case CDS_NO_INFO:
#ifdef LOG
      printf("no info\n");
#endif
      break;
    case CDS_NO_DISC:
#ifdef LOG
      printf("no disc\n");
#endif
      break;
    case CDS_TRAY_OPEN:
#ifdef LOG
      printf("tray open\n");
#endif
      break;
    case CDS_DRIVE_NOT_READY:
#ifdef LOG
      printf("drive not ready\n");
#endif
      break;
    case CDS_DISC_OK:
#ifdef LOG
      printf("disc ok\n");
#endif
      close(fd);
      return 1;
      break;
    default:
      printf("unknown value: %u\n", drive_status);
  }
  close(fd);
  return 0;
}

typedef struct cdrom_files_s cdrom_files_t;
static const struct cdrom_files_s {
  int type;
  char file[128];
} cdrom_files[] = {
  { AUTOCD_LOMB, "autorun.sh" },
  { AUTOCD_PSX, "system.cnf" },
  { AUTOCD_VCD, "vcd/info.vcd" },
  { AUTOCD_SVCD, "svcd/info.svd" },
  { AUTOCD_DVD, "audio_ts" },
  { AUTOCD_DVD, "video_ts" },
  { 0, "" }
};

int autocd_disc_type(const char *device, const char *mountpoint) {
  
  int type = AUTOCD_UNKNOWN;

  /*
   * stage 1: using CDROM_DISC_STATUS
   */
  {
    int fd;
    int disc_status;
    
    fd = open(device, O_RDONLY | O_NONBLOCK);
    if (fd<0) {
      printf("autocd: error opening device\n");
      return 0;
    }
    
    disc_status = ioctl (fd, CDROM_DISC_STATUS);
    
#ifdef LOG
    printf("autocd: disc status: ");
#endif

    switch(disc_status) {
      case CDS_AUDIO:
#ifdef LOG
	printf("audio\n");
#endif
	return AUTOCD_AUDIO;  /* finish here, got audio cd */
	break;
      case CDS_DATA_1:
#ifdef LOG
	printf("data 1\n");
#endif
	/* here: CD_ROM, mixed mode CD_ROM/CD_AUDIO, DVD_VIDEO, SVCD, ... ???? */
	type = AUTOCD_DATA1;
	break;
      case CDS_DATA_2:
#ifdef LOG
	printf("data 2\n");
#endif
	/* here: all XA based cd's. Can't guess anything here. */
	type = AUTOCD_DATA2;
	break;
      case CDS_XA_2_1:
#ifdef LOG
	printf("xa 1\n");
#endif
	/* here: XA based cd's */
	type = AUTOCD_XA1;
	break;
      case CDS_XA_2_2:
#ifdef LOG
	printf("xa 2\n");
#endif
	/* here: XA based cd's, no error correction: audio or video data... */
	type = AUTOCD_XA2;
	break;
      case CDS_MIXED:
#ifdef LOG
	printf("mixed\n");
#endif
	type = AUTOCD_MIXED;
	break;
      case CDS_NO_INFO:
#ifdef LOG
	printf("no info\n");
#endif
	break;
      case CDS_NO_DISC:
#ifdef LOG
	printf("no disc\n");
#endif
	type = AUTOCD_NODISC;
	break;
      default:
	printf("unknown value: %u\n", disc_status);
    }
    close(fd);
  }
  
  /*
   * stage 2: mounting cd and looking for specific files
   */
  {
    const cdrom_files_t *files = cdrom_files;
    char file[128];

    if (!autocd_mount(mountpoint)) return type;

    type = AUTOCD_CDROM;

    while (files->type) {
      snprintf(file, 127, "%s/%s", mountpoint, files->file);
      if (access(file, F_OK) >= 0) {
#ifdef LOG
        printf("autocd: %s found -> cd type %i\n", file, files->type);
#endif
        autocd_umount(mountpoint);
	return files->type;
      }
      files++;
    }
    autocd_umount(mountpoint);
  }
  
  /*
   * stage 3: raw reads
   */

  /* TODO */

  return type;
}


static char *get_biggest_file(const char *dir) {
  
  DIR *dirp;
  struct dirent *entp;
  char *file = NULL;
  off_t size = 0;
  
  dirp = opendir (dir);
  if (dirp != NULL)
    {
      entp = readdir (dirp);
      while (entp) {
	char *entry = ho_strdup(entp->d_name);
        struct stat filestat;
	char fullname[1024];
	
	snprintf(fullname, 1023, "%s/%s", dir, entry);
	stat(fullname, &filestat);

	if (filestat.st_size > size) {
	  if (file) ho_free(file);
	  file = ho_strdup(fullname);
	  size = filestat.st_size;
	}
	ho_free(entry);
	entp = readdir (dirp);
      }
      closedir (dirp);
    }
  else
    puts ("Couldn't open the directory.");

  return file;
}

void disc_eject(oxine_t *oxine) {

  int fd;
  oxine_event_t ev;
  
  ev.type = OXINE_EVENT_FINISH;

  if (oxine->cd_in_use) {
    odk_stop(oxine->odk);
    odk_send_event(oxine->odk, &ev);
  }
  if (oxine->cd_is_mounted) {
    autocd_umount(oxine->cd_mountpoint);
    oxine->cd_is_mounted = 0;
  }

  fd = open(oxine->cd_device, O_RDONLY | O_NONBLOCK);
  if (fd<0) {
    printf("autocd: error opening device\n");
    return;
  }
  
  ioctl (fd, CDROMEJECT);

  close(fd);
}

void disc_close(oxine_t *oxine) {

  int fd;
 
  fd = open(oxine->cd_device, O_RDONLY | O_NONBLOCK);
  if (fd<0) {
    printf("autocd: error opening device\n");
    return;
  }
  
  ioctl (fd, CDROMCLOSETRAY);

  close(fd);
}

static void disc_play_cb (void *data) {
  
  int type;
  oxine_t *oxine = (oxine_t*) data;
  otk_widget_t *b;

  if(!autocd_disc_in(oxine->cd_device)) {
    disc_cb(data);
    return;
  }

  printf("oxine: detecting ...\n");
  type = autocd_disc_type(oxine->cd_device,oxine->cd_mountpoint);
  printf("oxine: got: %u\n", type);

  
  oxine->main_window = NULL;
  otk_clear(oxine->otk);

  switch (type) {
    case AUTOCD_DVD:
      oxine->cd_in_use = 1;
      odk_pass_events(oxine->odk, 1);
      odk_set_mode(oxine->odk, ODK_MODE_NORM);
      odk_open(oxine->odk, "dvd://");
      odk_play(oxine->odk);
      break;
    case AUTOCD_VCD:
    case AUTOCD_SVCD:
      oxine->cd_in_use = 1;
      odk_pass_events(oxine->odk, 1);
      odk_set_mode(oxine->odk, ODK_MODE_NORM);
      odk_open(oxine->odk, "vcd://");
      odk_play(oxine->odk);
      break;
    case AUTOCD_CDROM:
      {
	char *file;
	
	if (autocd_mount(oxine->cd_mountpoint))
	{
	  oxine->cd_is_mounted = 1;
	  file = get_biggest_file(oxine->cd_mountpoint);
	  if(file) {
	    autocd_setspeed(oxine->cd_device, 4);
	    odk_open(oxine->odk, file);
	    odk_play(oxine->odk);
	    ho_free(file);
	  }
	}
        break;
      }
    case AUTOCD_AUDIO:
      oxine->cd_in_use = 1;
      odk_set_mode(oxine->odk, ODK_MODE_NORM);
      odk_open(oxine->odk, "cdda:1");
      odk_play(oxine->odk);
      break;
    default:
      oxine->main_window = otk_window_new (oxine->otk, NULL, 150, 200, 500, 200);
      otk_label_new(oxine->main_window, 250, 100, OTK_ALIGN_CENTER|OTK_ALIGN_BOTTOM, "disc currently unsupported");
      b = otk_button_new(oxine->main_window, 150, 140, 200, 50, "back", oxine->main_menu_cb, oxine);
      otk_set_focus(b);
      otk_draw_all(oxine->otk);
  }
}

void disc_cb (void *data) {

  oxine_t *oxine = (oxine_t*) data;
  otk_widget_t *b;
  
  if(autocd_disc_in(oxine->cd_device)) {
    disc_play_cb(data);
    return;
  } else {
    disc_close(oxine); /* try close first */
    if(autocd_disc_in(oxine->cd_device)) {
      disc_play_cb(data);
      return;
    }
    disc_eject(oxine); /* ok, open here */
  }
  
  oxine->main_window = NULL;
  otk_clear(oxine->otk);
  oxine->main_window = otk_window_new (oxine->otk, NULL, 200, 200, 400, 200);
  otk_label_new(oxine->main_window, 200, 100, OTK_ALIGN_CENTER|OTK_ALIGN_BOTTOM, "please insert a disc");
  b = otk_button_new(oxine->main_window, 220, 140, 160, 50, "done", disc_play_cb, oxine);
  otk_set_focus(b);
  otk_button_new(oxine->main_window, 20, 140, 160, 50, "back", oxine->main_menu_cb, oxine);
  
  otk_draw_all(oxine->otk);

}
