/*
 * Copyright (C) 2003 Stefan Holst
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: fb.c,v 1.1 2003/03/19 15:50:31 holstsn Exp $
 *
 * "window" for frame buffer devices
 */

#include <stdio.h>
#include <string.h>
#include <semaphore.h>
#include <xine.h>

#include "event.h"
#include "odk_plugin.h"
#include "otk.h"
#include "utils.h"

/*
#define LOG
*/

typedef struct {
  odk_window_t window_class;

  int width, height;
  double ratio;

  sem_t waiting;

  void (*event_handler)(void *data, oxine_event_t *ev);
  void *event_handler_data;

  fb_visual_t vis;
  
} fb_window_t;

static void frame_output_cb (void *data,
			     int video_width, int video_height,
			     double video_pixel_aspect,
			     int *dest_x, int *dest_y, 
			     int *dest_width, int *dest_height,
			     double *dest_pixel_aspect,
			     int *win_x, int *win_y) {

  fb_window_t *this = (fb_window_t *)data;
  static int width = 0;
  static int height = 0;
  
  if (this == NULL)
    return;

  if ((width != video_width) || (height != video_height)) {
    oxine_event_t ev;

    width = video_width;
    height = video_height;
#ifdef LOG
    printf("fb: video format changed to %ix%i\n", width, height);
#endif
	    
    ev.type = OXINE_EVENT_FORMAT_CHANGED;
    ev.x = width;
    ev.y = height;
    
    this->event_handler(this->event_handler_data, &ev);
  }

  *dest_x = 0;
  *dest_y = 0;

  *win_x = this->window_class.xpos;
  *win_y = this->window_class.ypos;

  *dest_width  = this->window_class.width;
  *dest_height = this->window_class.height;

}

/*
 * show - dummy function
 */
static int show(odk_window_t *this) {
  return 1;
}

/*
 * hide - dummy function
 */
static int hide(odk_window_t *this) {
  return 1;
}

/*
 * fullscreen - dummy function
 */
static int fullscreen(odk_window_t *this, int mode) {
  return 1;
}

/*
 * displace - alters geometry of window
 */
static int displace(odk_window_t *this, int x, int y, int w, int h) {
  fb_window_t *win = (fb_window_t *)this;
  win->width = w;
  win->height = h;
  return 1;
}

static void stop_event_loop(odk_window_t *this) {
  fb_window_t *win = (fb_window_t *)this;
  
  sem_post(&win->waiting);
}

/*
 * dispose - close window, free resources
 */
static int dispose(odk_window_t *this) {
  fb_window_t *win = (fb_window_t *)this;
  int are_waiting;
  sem_getvalue(&win->waiting, &are_waiting);
  if (are_waiting == 0) {
    stop_event_loop(this); /* let event_loop exit first */
  } else {
    xine_close_video_driver(this->xine, this->video_port);
    this->video_port = NULL;
    ho_free(win);
  }
  return 1;
}

/*
 * event_loop - just wait until the window is disposed
 */
static void event_loop(odk_window_t *this) {
  fb_window_t *win = (fb_window_t *)this;
  sem_wait(&win->waiting); /* wait for window disposal */
}

/*
 * set_event_handler - update event cb
 */
static void set_event_handler(odk_window_t *this,
			      void (*cb)(void *data, oxine_event_t *ev),
			      void *data) {
  fb_window_t *win = (fb_window_t *)this;
  win->event_handler = cb;
  win->event_handler_data = data;
}

/*
 * constructor
 */
odk_window_t *fb_construct(xine_t *xine, const char *driver) {
  fb_window_t *win = ho_new_tagged(fb_window_t, "fb window object");

  sem_init(&win->waiting, 0, 0);

  win->vis.frame_output_cb = frame_output_cb;
  win->vis.user_data = win;

  win->window_class.xine = xine;
  win->window_class.video_port = xine_open_video_driver(xine,
							"fb",
							XINE_VISUAL_TYPE_FB,
							(void*)&win->vis);
  win->window_class.show = show;
  win->window_class.hide = hide;
  win->window_class.displace = displace;
  win->window_class.dispose = dispose;
  win->window_class.fullscreen = fullscreen;
  win->window_class.event_loop = event_loop;
  win->window_class.stop_event_loop = stop_event_loop;
  win->window_class.set_event_handler = set_event_handler;
  
  return (odk_window_t *)win;
}

