/*
 * Copyright (C) 2002-2003 Stefan Holst
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: odk.c,v 1.43 2003/03/23 17:00:00 holstsn Exp $
 *
 * odk abstracts from xine and other software / hardware interfaces
 */

#include <stdio.h>
#include <string.h>
#include <xine.h>

#include "oxine.h"

#include "odk.h"
#include "odk_plugin.h"
#include "lirc.h"
#include "utils.h"

/*
 * plugin's header files. Add here for new plugins.
 */

#include "x11.h"
#include "nox.h"
#include "fb.h"

/*
 * available window plugin descriptions.
 * register new plugin descriptions from you plugin header file here.
 * These descriptions has to be defined in plugin's header file which
 * has to be included in (and only in) odk.c.
 */

#define WINDOW_PLUGINS { \
        x11_plugin_desc, \
	nox_plugin_desc, \
	fb_plugin_desc, \
	NULL }

#define V_WIDTH 800
#define V_HEIGHT 600

/* determined by xine engine: */
#define NUM_COLORS 256

#define MODE_MENU ((uint64_t)1 << 0)
#define MODE_OPEN ((uint64_t)1 << 1)
#define MODE_PLAY ((uint64_t)1 << 2)
#define MODE_GOOM ((uint64_t)1 << 3)
#define MODE_SEEK ((uint64_t)1 << 4)

#define PIXMAP_SIMPLE_ARROW_UP 1
#define PIXMAP_SIMPLE_ARROW_DOWN 2

/* Thank's to glib project */
#define CLAMP(x, low, high)  (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))

static pthread_t seek_thread;

struct seek_arg_t {
  odk_t *odk;
  int how;
};

/*
#define LOG
*/

struct  odk_s {

  xine_t     *xine;

  xine_osd_t *osd;
  double vscale, hscale;
  
  odk_window_t *win;

  void (*event_handler)(void *data, oxine_event_t *ev);
  void *event_handler_data;
  int pass_events; /* pass events to xine */

  uint32_t color[NUM_COLORS];
  uint8_t trans[NUM_COLORS];

  int palette_fill;
  char *menu_bg;

  int logo_mode;

  uint64_t mode;
  xine_post_t *goom;

  odk_input_t *lirc;
  pthread_mutex_t event_mutex;

  char   *current_mrl;
};


/* generates a gradient on palette out of given values */

static void palette_transition(odk_t *odk, int from_index, int to_index,
                               uint32_t from_color, uint8_t from_trans,
			       uint32_t to_color, uint8_t to_trans) {
  double cr_step, cb_step, y_step;
  uint8_t cr_from, cb_from, y_from;
  uint8_t cr_to, cb_to, y_to;
  
  double trans_step;
  int num;

  num = to_index - from_index;

  trans_step = (double)(to_trans - from_trans) / num;

  cb_from = from_color & 0xff;
  cr_from = (from_color >> 8) & 0xff;
  y_from  = (from_color >> 16) & 0xff;

  cb_to   = to_color & 0xff;
  cr_to   = (to_color >> 8) & 0xff;
  y_to    = (to_color >> 16) & 0xff;

  cb_step = (double)(cb_to - cb_from) / num;
  cr_step = (double)(cr_to - cr_from) / num;
  y_step  = (double)(y_to  - y_from) / num;

  for (num = from_index; num < to_index; num++) {
    uint8_t cb_cur = cb_from + (int8_t)(cb_step * (num-from_index));
    uint8_t cr_cur = cr_from + (int8_t)(cr_step * (num-from_index));
    uint8_t y_cur  =  y_from + (int8_t)( y_step * (num-from_index));
    
    odk->color[num] = cb_cur + (cr_cur<<8) + (y_cur<<16);
    odk->trans[num] = from_trans + trans_step * (num-from_index);
    /* printf("writing: color %x trans %u to %u\n", odk->color[num], odk->trans[num], num); */
  }
  odk->color[to_index] = to_color;
  odk->trans[to_index] = to_trans;
}

int odk_get_color(odk_t *odk, uint32_t color, uint8_t trans) {

  int i;

  for(i=0; i<odk->palette_fill; i++) {
    if ((odk->color[i] == color) && (odk->trans[i] == trans))
      return i;
  }
  if (odk->palette_fill == NUM_COLORS) {
    printf("odk: too many colors: palette is full\n");
    return 0;
  }
  odk->color[odk->palette_fill]=color;
  odk->trans[odk->palette_fill]=trans;

  xine_osd_set_palette(odk->osd, odk->color, odk->trans);

  odk->palette_fill++;

  return odk->palette_fill - 1;
}

int odk_alloc_text_palette(odk_t *odk, uint32_t fg_color, uint8_t fg_trans,
                           uint32_t bg_color, uint8_t bg_trans,
                           uint32_t bo_color, uint8_t bo_trans) {

  palette_transition(odk, odk->palette_fill, odk->palette_fill+5, 
      bg_color, bg_trans, bo_color, bo_trans);
  palette_transition(odk, odk->palette_fill+5, odk->palette_fill+9, 
      bo_color, bo_trans, fg_color, fg_trans);

  odk->palette_fill += 10;
  
  xine_osd_set_palette(odk->osd, odk->color, odk->trans);

  return odk->palette_fill-11;
}

static void odk_adapt(odk_t *odk);

/*
 * sends events to xine and to other registered
 * event handler
 */

static void odk_event_handler(void *this, oxine_event_t *ev) {

  odk_t *odk = (odk_t*) this;
  xine_event_t event;
  xine_input_data_t inp;

  event.stream = odk->win->stream;
  event.type = 0;

  pthread_mutex_lock(&odk->event_mutex);

  if (ev->type == OXINE_EVENT_FORMAT_CHANGED) {

    odk_adapt(odk);

  } else if (ev->type == OXINE_EVENT_MOTION) {

    event.type = XINE_EVENT_INPUT_MOUSE_MOVE;
    event.data = &inp;
    event.data_length = sizeof(inp);

    inp.button = 0;
    inp.x = ev->x;
    inp.y = ev->y;
  
  } else if (ev->type == OXINE_EVENT_BUTTON) {
    
    event.type = XINE_EVENT_INPUT_MOUSE_BUTTON;
    event.data = &inp;
    event.data_length = sizeof(inp);

    inp.button = ev->key;
    inp.x = ev->x;
    inp.y = ev->y;

  } else if (ev->type == OXINE_EVENT_KEY) {

    event.type = 0;
    event.data = NULL;
    event.data_length = 0;

    switch(ev->key) {

      case OXINE_KEY_MENU1:
	event.type = XINE_EVENT_INPUT_MENU1;
	break;
      case OXINE_KEY_MENU2:
	event.type = XINE_EVENT_INPUT_MENU2;
	break;
      case OXINE_KEY_MENU3:
	event.type = XINE_EVENT_INPUT_MENU3;
	break;
      case OXINE_KEY_MENU4:
	event.type = XINE_EVENT_INPUT_MENU4;
	break;
      case OXINE_KEY_MENU5:
	event.type = XINE_EVENT_INPUT_MENU5;
	break;
      case OXINE_KEY_MENU6:
	event.type = XINE_EVENT_INPUT_MENU6;
	break;
      case OXINE_KEY_UP:
	event.type = XINE_EVENT_INPUT_UP;
	break;
      case OXINE_KEY_DOWN:
	event.type = XINE_EVENT_INPUT_DOWN;
	break;
      case OXINE_KEY_LEFT:
	event.type = XINE_EVENT_INPUT_LEFT;
	break;
      case OXINE_KEY_RIGHT:
	event.type = XINE_EVENT_INPUT_RIGHT;
	break;
      case OXINE_KEY_SELECT:
	event.type = XINE_EVENT_INPUT_SELECT;
	break;
      case OXINE_KEY_NEXT:
	event.type = XINE_EVENT_INPUT_NEXT;
	break;
      case OXINE_KEY_PRIOR:
	event.type = XINE_EVENT_INPUT_PREVIOUS;
	break;
      case OXINE_KEY_0:
	event.type = XINE_EVENT_INPUT_NUMBER_0;
	break;
      case OXINE_KEY_1:
	event.type = XINE_EVENT_INPUT_NUMBER_1;
	break;
      case OXINE_KEY_2:
	event.type = XINE_EVENT_INPUT_NUMBER_2;
	break;
      case OXINE_KEY_3:
	event.type = XINE_EVENT_INPUT_NUMBER_3;
	break;
      case OXINE_KEY_4:
	event.type = XINE_EVENT_INPUT_NUMBER_4;
	break;
      case OXINE_KEY_5:
	event.type = XINE_EVENT_INPUT_NUMBER_5;
	break;
      case OXINE_KEY_6:
	event.type = XINE_EVENT_INPUT_NUMBER_6;
	break;
      case OXINE_KEY_7:
	event.type = XINE_EVENT_INPUT_NUMBER_7;
	break;
      case OXINE_KEY_8:
	event.type = XINE_EVENT_INPUT_NUMBER_8;
	break;
      case OXINE_KEY_9:
	event.type = XINE_EVENT_INPUT_NUMBER_9;
	break;
    }
  } else return;

  if (event.type && odk->pass_events)
    xine_event_send (odk->win->stream, &event);

  odk_send_event(odk, ev);
  pthread_mutex_unlock(&odk->event_mutex);
}


/*
 * adapt osd to new stream size.
 * This only affects primitives drawn after
 * this call.
 */

static void odk_adapt(odk_t *odk) {

  int width, height, ratio, format;

  if (odk->osd) xine_osd_free(odk->osd);

  if (!xine_get_current_frame (odk->win->stream, &width, &height, &ratio, &format, NULL))
  {
#ifdef LOG
    printf("odk: odk_adapt: xine_get_current_frame failed\n");
#endif
    width       = xine_get_stream_info(odk->win->stream, XINE_STREAM_INFO_VIDEO_WIDTH);
    height      = xine_get_stream_info(odk->win->stream, XINE_STREAM_INFO_VIDEO_HEIGHT);
  }

#ifdef LOG
  printf("odk: odk_adapt to: %i, %i\n", width, height);
#endif
  odk->vscale = height / (double)V_HEIGHT;
  odk->hscale = width / (double)V_WIDTH;
  
  odk->osd = xine_osd_new(odk->win->stream, 0, 0, width, height);
  xine_osd_set_palette(odk->osd, odk->color, odk->trans);
}


uint8_t *odk_get_pixmap(int type) {
  int i;
  uint8_t arrowup[] = {0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 
		       0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		       0, 0, 1, 0, 1, 1, 0, 1, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
		       0, 0, 0, 0, 1, 1, 0, 0, 0, 0};
  uint8_t arrowdown[] = {0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 
			 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
			 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
			 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
			 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
			 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
			 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
			 0, 0, 1, 0, 1, 1, 0, 1, 0, 0,
			 0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
			 0, 0, 1, 1, 1, 1, 1, 1, 0, 0};

  uint8_t *p = NULL;

  p = ho_newstring(10*10 + 10);

  switch(type) {
  case PIXMAP_SIMPLE_ARROW_UP:
    memcpy(p, arrowup, 100);
    break;
  case PIXMAP_SIMPLE_ARROW_DOWN:
    memcpy(p, arrowdown, 100);
    break;
  }

  for(i=0;i<100;i++) p[i] = p[i] * XINE_OSD_TEXT1 + 2;

  return p;
}


/*
 * xine event thread
 */

static void xine_event_cb (void *user_data, const xine_event_t *event) {

  odk_t *odk = (odk_t *)user_data;
  oxine_event_t ev;
  
  switch (event->type) {
  case XINE_EVENT_UI_PLAYBACK_FINISHED:
    if (!odk->logo_mode) {
      ev.type = OXINE_EVENT_FINISH;
      odk_send_event(odk, &ev);
    }
    break;
  }
}

/*
 * initializes drawable, xine stream and osd
 */

odk_t *odk_init(xine_t *xine) {

  window_plugin_desc_t *desc[] = WINDOW_PLUGINS;
  window_plugin_desc_t *desc_ptr;
  odk_t *odk = ho_new_tagged(odk_t, "odk object");
  xine_event_queue_t *queue;
  xine_cfg_entry_t centry;
  int i=1;

#ifdef LOG
  printf("odk: initializing\n");
#endif
  
  odk->palette_fill = 1;

  odk->xine = xine;
  pthread_mutex_init(&odk->event_mutex, NULL);

  xine_config_lookup_entry (xine, "oxine.preferred_video", &centry);

  /* try to initialize a window plugin which is capable to handle
     our video driver */
  i=0;
  desc_ptr=desc[i];
  odk->win = NULL;
  while(desc_ptr) {
    while (desc_ptr->driver && strcasecmp(desc_ptr->driver, centry.str_value)) 
      desc_ptr++;
    if (desc_ptr->driver) {
      printf("odk: using %s\n", desc_ptr->desc);
      odk->win = desc_ptr->constructor(xine, desc_ptr->driver);
      if (odk->win) break;
    }
    desc_ptr=desc[++i];
  }
  
  if (!odk->win) {
    printf("odk: odk_init: error while initializing drawable\n");
    return NULL;
  }
  
  odk->win->set_event_handler(odk->win, odk_event_handler, odk);

  xine_config_lookup_entry (xine, "oxine.preferred_audio", &centry);

  /* create audio port */
  odk->win->audio_port = xine_open_audio_driver (xine, centry.str_value, NULL);
  if (!odk->win->audio_port) {
    printf("odk: odk_init: failed to open audio port '%s', falling back to auto\n", centry.str_value);
    odk->win->audio_port = xine_open_audio_driver (xine, NULL, NULL);
  }
  
  /* create stream */
  odk->win->stream = xine_stream_new(xine, odk->win->audio_port, odk->win->video_port);
  
  /* event handling */
  queue = xine_event_new_queue (odk->win->stream);
  xine_event_create_listener_thread (queue, xine_event_cb, odk);

  odk_adapt(odk);

  return odk;
}


/*
 * primitive drawing functions
 */

void odk_draw_bitmap(odk_t *odk, uint8_t *bitmap, int x1, int y1, int width, int height,
		     uint8_t *palette_map)
{
  /* uint8_t a[] = {0, 0, 0, 255, 255, 255}; */
  xine_osd_draw_bitmap(odk->osd, bitmap, x1, y1, width, height, NULL);

}


void odk_draw_line(odk_t *odk, int x1, int y1, int x2, int y2, int color) {

 int px1, py1, px2, py2;  

 px1 = (int) ((double)x1*odk->hscale);
 py1 = (int) ((double)y1*odk->vscale);
 px2 = (int) ((double)x2*odk->hscale);
 py2 = (int) ((double)y2*odk->vscale);
#ifdef LOG
  printf("odk: drawing line %u %u %u %u col: %u\n", px1, py1, px2, py2, color);
#endif
      
  xine_osd_draw_line(odk->osd, 
      px1, 
      py1, 
      px2, 
      py2, 
      XINE_OSD_TEXT1 + color);
}


void odk_draw_rect(odk_t *odk, int x1, int y1, int x2, int y2, int filled, int color) {

 int px1, py1, px2, py2;  

 px1 = (int) ((double)x1*odk->hscale);
 py1 = (int) ((double)y1*odk->vscale);
 px2 = (int) ((double)x2*odk->hscale);
 py2 = (int) ((double)y2*odk->vscale);
#ifdef LOG
  printf("odk: drawing rect %u %u %u %u col: %u\n", px1, py1, px2, py2, color);
#endif
      
  xine_osd_draw_rect(odk->osd, 
      px1, py1, px2, py2, XINE_OSD_TEXT1 + color, filled);
}


void odk_draw_text(odk_t *odk, int x, int y, const char *text, int alignment, int color) {

  int px, py, w, h;
  
  odk_get_text_size(odk, text, &w, &h);

  if (alignment & ODK_ALIGN_CENTER)  x -= w/2;
  if (alignment & ODK_ALIGN_RIGHT)   x -= w;
  if (alignment & ODK_ALIGN_VCENTER) y -= h/2;
  if (alignment & ODK_ALIGN_BOTTOM)  y -= h;
 
  px = (int) ((double)x*odk->hscale);
  py = (int) ((double)y*odk->vscale);

#ifdef LOG
  printf("odk: drawing text at %u %u (vs: %f hs: %f)\n", px, py, odk->vscale, odk->hscale);
#endif
  
  xine_osd_draw_text(odk->osd, px, py, text, color);
}


/*
 * overall osd control
 */

void odk_show(odk_t *odk) {
  xine_osd_show(odk->osd, 0);
} 

void odk_hide(odk_t *odk) {
  xine_osd_hide(odk->osd, 0);
} 

void odk_clear(odk_t *odk) {
  xine_osd_clear(odk->osd);
} 


/*
 * font stuff
 */

void odk_get_text_size(odk_t *odk, const char *text, int *width, int *height) {

  int w, h;

  xine_osd_get_text_size(odk->osd, text, &w, &h);
  *width = (int) ((double)w/odk->hscale);
  *height = (int) ((double)h/odk->vscale);
}


void odk_set_font(odk_t *odk, const char *font, int size) {

  int psize;

  psize = (int)((double)size*(odk->hscale+odk->vscale)/2);

  /* smallest text size possible */
  /* printf("%d\n", psize); */
  if(psize) 
    if (psize<16) psize=16;
  
#ifdef LOG
  printf("odk: setting font to %s %u %u\n", font, size, psize);
#endif
  
  xine_osd_set_font(odk->osd, font, psize);
}


/*
 * event stuff
 */

void odk_set_event_handler(odk_t *odk, void (*cb)(void *data, oxine_event_t *ev), void *data) {

  odk->event_handler = cb;
  odk->event_handler_data = data;
}

void odk_send_event(odk_t *odk, oxine_event_t *event) {

  switch (event->type) {
    case OXINE_EVENT_BUTTON:
    case OXINE_EVENT_MOTION:
      event->x = (int) ((double)event->x/odk->hscale);
      event->y = (int) ((double)event->y/odk->vscale);
      if ((event->x < 0) || (event->x > V_WIDTH)) return;
      if ((event->y < 0) || (event->y > V_HEIGHT)) return;
  }
  if (odk->event_handler)
    odk->event_handler(odk->event_handler_data, event);

}

void odk_pass_events(odk_t *odk, int mode) {

  odk->pass_events = mode;
}

/*
 * destructor
 */

void odk_free(odk_t *odk) {

#ifdef LOG
  printf("odk: finalizing\n");
#endif

  if (odk->current_mrl) ho_free(odk->current_mrl);

  xine_osd_free(odk->osd);
  xine_dispose (odk->win->stream);
  xine_close_audio_driver (odk->xine, odk->win->audio_port);
  
  if (odk->lirc)
    stop_lirc(odk->lirc);
  odk->win->dispose(odk->win);

  ho_free(odk);
}


/*
 * xine control
 */

/*
 * stops a running stream (if there is one),
 * opens a new stream, returns 0 on success.
 */

int odk_open(odk_t *odk, const char *mrl) {

  int ret;
  char error[255];
  
  if (odk->mode & MODE_PLAY) {
    odk_stop(odk);
  }

#ifdef LOG
  printf("odk: odk_open: trying to open '%s'\n", mrl);
#endif
  
  if (odk->current_mrl) ho_free(odk->current_mrl);
  ret = xine_open(odk->win->stream, mrl);
  
  if(!ret) {
    ret = xine_get_error(odk->win->stream);
    switch(ret) {
      case XINE_ERROR_NO_INPUT_PLUGIN:
	strncpy(error, "No input plugin", 255);
	break;
      case XINE_ERROR_NO_DEMUX_PLUGIN:
	strncpy(error, "No demux plugin", 255);
	break;
      case XINE_ERROR_DEMUX_FAILED:
	strncpy(error, "Demux failed", 255);
	break;
      case XINE_ERROR_MALFORMED_MRL:
	strncpy(error, "Malformed mrl", 255);
	break;
      default:
	strncpy(error, "Unknown error", 255);
	break;
    }
    printf("odk: odk_open: error %d (%s)\n", ret, error);
  } else {
#ifdef LOG
    printf("odk: odk_open: success.\n");
#endif
    odk->mode |= MODE_OPEN;

    odk->current_mrl = ho_strdup(mrl);

    odk_adapt(odk);

    return 0;
  }
 

  return ret;
}


void odk_play(odk_t *odk) {

  int ret;
  int i = 0, hasvideo;
  xine_post_in_t *input;
  xine_post_out_t *audio_source;
  oxine_event_t *ev;

  if(!(odk->mode & MODE_OPEN)) {
    printf("odk: odk_play: no stream opened\n");
    return;
  }
  
    hasvideo = xine_get_stream_info(odk->win->stream, XINE_STREAM_INFO_HAS_VIDEO);

  if(!hasvideo) {
    
    odk->goom = xine_post_init(odk->xine, "goom", 0, &odk->win->audio_port, &odk->win->video_port);
    if(odk->goom) {
      audio_source = xine_get_audio_source(odk->win->stream);
      input = (xine_post_in_t *) xine_post_input(odk->goom, "audio in");

      i = xine_post_wire(audio_source, input);

      odk->mode |= MODE_GOOM;
      }

  } 

  odk->mode |= MODE_PLAY;
  odk->mode &= ~MODE_SEEK;
  ret = xine_play(odk->win->stream, 0, 0);
  if(!ret) {
    ret = xine_get_error(odk->win->stream);
    printf("odk: odk_play: got error code: %d\n", ret);
    return ;
  }
  ev = ho_new(oxine_event_t);
  ev->type = OXINE_EVENT_START;
  odk_send_event(odk, ev);
}

void odk_stop(odk_t *odk) {

  xine_post_out_t *pp;

  xine_close(odk->win->stream);
  if(odk->mode & MODE_GOOM) {
    pp = xine_get_audio_source(odk->win->stream);
    xine_post_wire_audio_port(pp, odk->win->audio_port);
    xine_post_dispose(odk->xine, odk->goom);
    odk->mode &= ~MODE_GOOM;
  }
  odk->mode &= ~MODE_PLAY;
}

int get_pos_length(xine_stream_t *stream, int *pos, int *time, int *length) {                                   
  int t = 0, ret = 0;
  
  if(stream && (xine_get_status(stream) == XINE_STATUS_PLAY)) {
    while(((ret = xine_get_pos_length(stream, pos, time, length)) == 0) && (++t < 10))
      /*printf("wait");*/
      usleep(100000); /* wait before trying again */
  }
  return ret;
}   

int odk_get_pos_length_high(odk_t *odk, int *pos, int *time, int *length) {
  int ret = 0;
  static int last_time = 0;
  xine_stream_t *stream = odk->win->stream;

  if(stream && (xine_get_status(stream) == XINE_STATUS_PLAY)) {                                                          
    ret = xine_get_pos_length(stream, pos, time, length);
  }

  /*
   * smoothen the times xine returns a bit. filtering out small backjumps here.
   */
  
  if (ret) {
    if ((*time < last_time) && (*time+1000 > last_time)) {
      *time = last_time;
    } else {
      last_time = *time;
    }
  }

  return ret;  
} 

int odk_get_pos_length(odk_t *odk, int *pos, int *time, int *length) {
  return get_pos_length(odk->win->stream, pos, time, length);
}

/* Return stream position 1..100 */
int odk_get_seek(odk_t *odk) {
  int pos_time, length;
  int pos=1;
  if(!odk_get_pos_length_high(odk, NULL, &pos_time, &length)) {
    return -1;
  }
  if (length) pos = ((pos_time*100) / length);
  /* printf("seek pos : %d\n", pos); */
  return pos;
}

/* Thank's to xine-ui project */
void *seek_relative_thread(void *data) {                                                                             
  struct seek_arg_t *data1 = (struct seek_arg_t *)data;
  int how = (int)data1->how; 
  int pos_time, length;
  int final;        
  odk_t *odk = data1->odk;   

  pthread_detach(pthread_self());                                                                                        

  if(!get_pos_length(odk->win->stream, NULL, &pos_time, &length)) {
    /* printf("Not ret\n");*/   
    ho_free(data1);
    odk->mode &= ~MODE_SEEK;
    pthread_exit(NULL);   
    return NULL;                                                                                              
  }    

  pos_time /= 1000;
                                                                        
  final = pos_time + how;
  /*printf("final, len : %d %d \n", CLAMP(final*1000, 0, (length-2000)), length);*/
  xine_play(odk->win->stream, 0, CLAMP(final*1000, 0, (length-2000)));
  /* printf("oi\n");  */
  /*  (void) xine_play(odk->win->stream, 0, final); */
  ho_free(data1);          
  odk->mode &= ~MODE_SEEK;                                                                                 
  pthread_exit(NULL);                                                                                                
  return NULL;                                                                                                           
}         

void odk_seek(odk_t *odk, int how) {
  int seekable, err;
  struct seek_arg_t *pas;

  if(odk->mode & MODE_SEEK){
    /*printf("Already seeking\n");*/
    return;
  }
  if(odk->logo_mode == 1) return;
  seekable = xine_get_stream_info(odk->win->stream, XINE_STREAM_INFO_SEEKABLE);
  if(!seekable) {
    printf("Video is not seekable\n");
    return;
  }
  /*printf("seek %d\n", how);*/
  pas = ho_new(struct seek_arg_t);
  pas->odk = odk;
  pas->how = how;
  odk->mode |= MODE_SEEK;                                                                                   
  if ((err = pthread_create(&seek_thread,                                                                                
                            NULL, seek_relative_thread, (void *)pas)) != 0) {
    printf("can't create new thread\n");
    ho_free(pas);   
    odk->mode &= ~MODE_SEEK;
    abort();                                                                                                             
  }   
}

void odk_set_speed(odk_t *odk, int speed) {

  int s = XINE_SPEED_NORMAL;
  
  switch(speed) {
    case ODK_SPEED_PAUSE:
      s = XINE_SPEED_PAUSE;
      break;
    case ODK_SPEED_NORMAL:
      s = XINE_SPEED_NORMAL;
      break;
    default:
      printf("odk: odk_set_speed: invalid speed\n");
  }
  xine_set_param(odk->win->stream, XINE_PARAM_SPEED, s);
}

int odk_get_speed(odk_t *odk) {

  int s=xine_get_param(odk->win->stream, XINE_PARAM_SPEED);
  
  switch(s) {
    case XINE_SPEED_PAUSE:
      return ODK_SPEED_PAUSE;
      break;
    case XINE_SPEED_NORMAL:
      return ODK_SPEED_NORMAL;
      break;
  }
  return -1;
}

void odk_toggle_pause(odk_t *odk) {

  int speed = xine_get_param(odk->win->stream, XINE_PARAM_SPEED);
  int speed2;

  switch(speed) {
  case XINE_SPEED_PAUSE:
    speed2 = XINE_SPEED_NORMAL;
    break;
  default:
    speed2 = XINE_SPEED_PAUSE;
    break;
  }

  xine_set_param(odk->win->stream, XINE_PARAM_SPEED, speed2);
}

void odk_eject(odk_t *odk) {

  xine_eject(odk->win->stream);
}

void odk_volmute(odk_t *odk) {

  int mute =  xine_get_param(odk->win->stream, XINE_PARAM_AUDIO_MUTE);
  xine_set_param(odk->win->stream, XINE_PARAM_AUDIO_MUTE, 1-mute);
}

void odk_set_volume(odk_t *odk, int how) {
  static int vol = -1; 
  int volume;

  /*printf("%d\n", vol);*/
  if(vol == -1) vol = xine_get_param(odk->win->stream, XINE_PARAM_AUDIO_VOLUME) * 10;
  vol += 7*how;
  if((vol <= 0)||(vol >= 1000)) {
    vol -= 7*how;
    return;
  }
  volume = vol / 10;  
  xine_set_param(odk->win->stream, XINE_PARAM_AUDIO_VOLUME, volume);
   /*  printf("volume %d\n", volume);*/
}

int odk_get_volume(odk_t *odk) {
  int volume;
  int mute =  xine_get_param(odk->win->stream, XINE_PARAM_AUDIO_MUTE);
  if(mute) return 0;
  volume =  xine_get_param(odk->win->stream, XINE_PARAM_AUDIO_VOLUME);
  if(volume < 0) return 0;
  if(volume > 100) return 100;
  return volume;
}

char *odk_get_mrl(odk_t *odk) {
  return odk->current_mrl;
}

/*
 * window control
 */

void odk_show_window(odk_t *odk) {
  odk->win->show(odk->win);
}

void odk_hide_window(odk_t *odk) {
  odk->win->hide(odk->win);
}

void odk_run(odk_t *odk) {
  odk->win->event_loop(odk->win);
}

void odk_exit(odk_t *odk) {
  odk->win->stop_event_loop(odk->win);
}

void odk_fullscreen(odk_t *odk, int mode) {
  odk->win->fullscreen(odk->win, mode);
}

/*
 * misc
 */

void odk_listen_to(odk_t *odk, const char *id) {

  if (!strcmp(id,"lirc")) {
    if (!odk->lirc)
      odk->lirc=start_lirc(odk, odk_event_handler);
  }
}

void odk_set_mode(odk_t *odk, int mode) {

  if (mode == ODK_MODE_LOGO) {
#ifdef LOG
    printf("odk: logo mode\n");
#endif
    odk->logo_mode = 1;
  } else {
#ifdef LOG
    printf("odk: normal mode\n");
#endif
    odk->logo_mode = 0;
  }
}

int odk_is_logo_mode(odk_t *odk) {
  return odk->logo_mode;
}

char *odk_get_meta_info(odk_t *odk, int info) {
  const char *str = xine_get_meta_info(odk->win->stream, info);

  if (!str) return NULL;
  return ho_strdup(str);
}

void odk_user_color(odk_t *odk, char *name, uint32_t *color, uint8_t *trans) {

  char id[512];
  char value[64];
  const char *v;
  unsigned int c, t;

  snprintf(id, 511, "oxine.color_%s", name);
  snprintf(value, 63, "%06x-%01x", *color, *trans);
  v = xine_config_register_string (odk->xine, id, value,
				"color specification yuv-opacity", 
				"color spec. format: [YUV color]-[opacity 0-f]",
				10, NULL, NULL);
  if (sscanf(v, "%06x-%01x", &c, &t) != 2) {
    printf("odk: bad formated color spec in entry '%s'\n", id);
    printf("odk: using standard color\n");
    return;
  }
  *color = c;
  *trans = t;
}
