/*
 * Copyright (C) 2002 Stefan Holst
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: playlist.c,v 1.9 2003/03/23 16:39:40 holstsn Exp $
 *
 * a simple playlist implementataion
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "playlist.h"
#include "list.h"
#include "oxine.h"
#include "utils.h"

/*
#define LOG
*/

/*
 * private data structs
 */

typedef struct play_item_s play_item_t;

struct play_item_s {

  char   *title;
  char   *mrl;

  int     played; /* bool indicating this has been played in playlist */
};

struct playlist_s {

  list_t *items; /* of play_item_t */

  int length;
  int selected;
};

/*
 * create an empty playlist
 */

playlist_t *playlist_new() {

  playlist_t *playlist;

  playlist = ho_new(playlist_t);

  playlist->items = list_new();
  playlist->length = 0;
  playlist->selected = -1;

  return playlist;
}

/*
 * append item to playlist, deep copies are made from mrl and title.
 */

void playlist_add(playlist_t *playlist, const char *mrl, const char *title) {

  play_item_t *p;

  if (!mrl) return;
  
#ifdef LOG
  printf("playlist: adding mrl: '%s' title: '%s'\n", mrl, title);
#endif
  p = ho_new(play_item_t);

  p->mrl = ho_strdup(mrl);
  if (title) 
    p->title = ho_strdup(title);
  else
    p->title = ho_strdup(mrl);

  p->played = 0;

  list_append_content(playlist->items, p);

  playlist->length++;
}

/*
 * remove all items from playlist
 */

void playlist_clear(playlist_t *playlist) {

  play_item_t *p;

  p = list_first_content(playlist->items);

  while (p) {
    if (p->mrl) ho_free(p->mrl);
    if (p->title) ho_free(p->title);
    ho_free(p);
    list_delete_current(playlist->items);
    p = list_first_content(playlist->items);
  }

  playlist->length = 0;
  playlist->selected = -1;
}

/*
 * get first unplayed item from list and mark it as played(and selected)
 * returns 0 if no unplayed items are available, 1 otherwise.
 */

int playlist_get_first_unplayed(playlist_t *playlist, char **mrl, char **title) {

  play_item_t *p;
  int i=0;

  p = list_first_content(playlist->items);
  while (p && p->played) {
    p = list_next_content(playlist->items);
    i++;
  }

  if (!p) return 0;

  *mrl = p->mrl;
  *title = p->title;
  p->played = 1;
  printf("pl selecting ... %d\n", i);
  playlist->selected = i;

  return 1;
}

int playlist_get_next(playlist_t *playlist, char **mrl, char **title)
{
  int i;
  i = playlist_get_current(playlist);
  i+=1;
  if(i > playlist->length) i=0;
  playlist->selected = i;

  return playlist_get_nth(playlist, i, mrl, title);
}

int playlist_get_prev(playlist_t *playlist, char **mrl, char **title)
{
  int i;
  i = playlist_get_current(playlist);
  i-=1;
  if(i==0) i=playlist->length;
  playlist->selected = i;

  return playlist_get_nth(playlist, i, mrl, title);
}

int playlist_get_current(playlist_t *playlist) {
  return playlist->selected;
}

/*
 * mark all items as unplayed
 */

void playlist_reset(playlist_t *playlist) {

  play_item_t *p;

  p = list_first_content(playlist->items);
  while (p) {
    p->played = 0;
    p = list_next_content(playlist->items);
  }
}

/*
 * get length of playlist (played and unplayed items are counted)
 */

int playlist_length(playlist_t *playlist) {

  return playlist->length;
}

/*
 * get nth item
 * returns 1 on success, 0 on error.
 */

int playlist_get_nth(playlist_t *playlist, int n, char **mrl, char **title) {

  play_item_t *p;
  int i;

  p = list_first_content(playlist->items);
  
  if (!p) return 0;
  
  for (i=1; i<n; i++) {
    p = list_next_content(playlist->items);
    if (!p) return 0;
  }

  *mrl = p->mrl;
  *title = p->title;

  return 1;
}

/*
 * remove nth item from playlist
 * returns 1 on success, 0 on error.
 */

int playlist_remove_nth(playlist_t *playlist, int n) {

  play_item_t *p;
  int i;

  p = list_first_content(playlist->items);
  
  if (!p) return 0;
  
  for (i=1; i<n; i++) {
    p = list_next_content(playlist->items);
    if (!p) return 0;
  }

  if (p->mrl) ho_free(p->mrl);
  if (p->title) ho_free(p->title);
  ho_free(p);
  list_delete_current(playlist->items);

  if(playlist->selected > n) playlist->selected -= 1;
  playlist->length--;
  
  return 1;
}

/*
 * frees given playlist and all its items
 */

void playlist_free(playlist_t *playlist) {

  playlist_clear(playlist);
  list_free(playlist->items);
  ho_free(playlist);
}

/*
 * playlist gui
 */

#define PL_ACTION_PLAY 1
#define PL_ACTION_REMOVE 2

const char *pl_action_strings[] = { "Play", "Remove" };

typedef struct {
  oxine_t *oxine;
  int action;
  int listpos;
  otk_widget_t *list;

} pl_session_t;

static void playlist_reentry_cb (void *data);

static void changelist (otk_widget_t *list, playlist_t *playlist) {

  play_item_t *item;
  list_t *items = playlist->items;
  char tmp[512];

  otk_remove_listentries(list);

  item = list_first_content(items);

  while (item) {
    if (item->played)
      snprintf(tmp, 511, "[p] %s", item->title);
    else
      snprintf(tmp, 511, "[u] %s", item->title);
    otk_add_listentry(list, tmp, item, -1);
    item = list_next_content(items);
  }
  otk_list_set_pos(list,0);
  otk_set_focus(list);
}

static void playlist_play_cb(void *data, void *entry) {

  pl_session_t *session = (pl_session_t*) data;
  oxine_t *oxine = session->oxine;
  play_item_t *item = (play_item_t*) entry;
  int n = otk_list_get_entry_number(session->list);

  session->listpos = otk_list_get_pos(session->list);

  if (session->action == PL_ACTION_PLAY) {
    oxine->playlist->selected = n;
    printf("playlist: playing %s n%d\n", item->mrl, n);    
    if(!odk_open(oxine->odk, item->mrl)) {
      item->played = 1;
      oxine->main_window = NULL;
      otk_clear(oxine->otk);
      odk_set_mode(oxine->odk, ODK_MODE_NORM);
      odk_play(oxine->odk);      
    }
  }
  if (session->action == PL_ACTION_REMOVE) {
    playlist_remove_nth(oxine->playlist, n);
    playlist_reentry_cb(session);
  }
}

static void playlist_clear_cb (void *data) {

  pl_session_t *session = (pl_session_t*) data;

  playlist_clear(session->oxine->playlist);

  playlist_reentry_cb(data);
}

static void playlist_reset_cb (void *data) {

  pl_session_t *session = (pl_session_t*) data;

  playlist_reset(session->oxine->playlist);

  playlist_reentry_cb(data);
}

static void playlist_leave_cb (void *data) {

  pl_session_t *session = (pl_session_t*) data;

  session->oxine->reentry = NULL;
  session->oxine->reentry_data = NULL;
  
  session->oxine->main_menu_cb(session->oxine);
}

static void playlist_leaveplay_cb (void *data) {

  pl_session_t *session = (pl_session_t*) data;
  oxine_t *oxine = session->oxine;
  char *mrl, *title;

  session->oxine->reentry = NULL;
  session->oxine->reentry_data = NULL;

  if (playlist_get_first_unplayed(session->oxine->playlist, &mrl, &title)) {
    if(!odk_open(oxine->odk, mrl)) {
      oxine->main_window = NULL;
      otk_clear(oxine->otk);
      odk_set_mode(oxine->odk, ODK_MODE_NORM);
      odk_play(oxine->odk);      
    }
  }
}

static void action_changed (void *data, int pos) {

  pl_session_t *session = (pl_session_t*) data;

  session->action = pos;
}

static void playlist_reentry_cb (void *data) {

  pl_session_t *session = (pl_session_t*) data;
  oxine_t      *oxine = session->oxine;
  otk_widget_t *b, *list_window;

  lock_job_mutex();
  if (oxine->blending_job) {
    cancel_job(oxine->blending_job);
  }
  if (oxine->info_window && oxine->media_info_close_cb) {
    oxine->media_info_close_cb(oxine);
  }
  unlock_job_mutex();

  oxine->main_window = NULL;
  otk_clear(oxine->otk);

  oxine->main_window = otk_window_new (oxine->otk, NULL, 20, 140, 200, 420);

  otk_label_new (oxine->main_window, 100, 60, OTK_ALIGN_CENTER|OTK_ALIGN_BOTTOM, "Action:");
  b = otk_selector_new (oxine->main_window, 10, 80, 180, 60, pl_action_strings, 2, action_changed, session);
  otk_set_focus(b);
  otk_selector_set(b, session->action);
  otk_button_new (oxine->main_window, 10, 150, 180, 60, "Play all", playlist_leaveplay_cb, session);
  otk_button_new (oxine->main_window, 10, 210, 180, 60, "Reset", playlist_reset_cb, session);
  otk_button_new (oxine->main_window, 10, 270, 180, 60, "Clear", playlist_clear_cb, session);
  otk_button_new (oxine->main_window, 10, 340, 180, 60, "Main Menu", playlist_leave_cb, session);

  list_window = otk_window_new (oxine->otk, NULL, 250, 140, 530, 420);

  session->list = otk_list_new(list_window, 10, 15, 510, 390, playlist_play_cb, session); 
  changelist(session->list, session->oxine->playlist);
  otk_list_set_pos(session->list, session->listpos);
  
  otk_draw_all(oxine->otk);
}


void playlist_cb (void *data) {

  oxine_t      *oxine = (oxine_t*) data;
  pl_session_t *session;

  session = ho_new(pl_session_t);
  session->oxine = oxine;
  session->action = 1;

  oxine->reentry_data = session;
  oxine->reentry = playlist_reentry_cb;

  playlist_reentry_cb(session);
}
