/*
 * Copyright (C) 2002-2003 Stefan Holst
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: x11.c,v 1.17 2003/03/23 18:48:52 holstsn Exp $
 *
 * the whole x11 stuff
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include <X11/extensions/XShm.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#ifdef HAVE_XTESTEXTENSION
#include <X11/extensions/XTest.h>
#endif /* HAVE_XTESTEXTENSION */


#include <xine.h>

#include "odk_plugin.h"
#include "event.h"
#include "utils.h"

#include "config.h"

/*
#define LOG
*/

#ifndef XShmGetEventBase
extern int XShmGetEventBase(Display *);
#endif

typedef struct {

  odk_window_t  window_class;

  Display        *display;
  int             screen;
  double          display_ratio;
  int             completion_event;
  Atom            wm_del_win;
  
  Window          window;
  Window          fullscreen_window;
  x11_visual_t    vis;

  Cursor          no_cursor;
  Cursor          on_cursor;
  int             cursor_visible;
  int             switchoff_job;
  
  int             sav_x, sav_y, sav_w, sav_h;
  int             display_width, display_height;
  int             fullscreen_mode;

  int             keep_going;

  odk_cb   event_handler;
  void *event_handler_data;

  /* screensaver disable */
  int             have_xtest;
  int             xtest_keycode;
  int             keep_screensaver;
  int             screensaver_job;

} x11_window_t;

/*
 * callbacks
 */

static void dest_size_cb (void *data,
			  int video_width, int video_height,
			  double video_pixel_aspect,
			  int *dest_width, int *dest_height,
			  double *dest_pixel_aspect) {
  
  x11_window_t *this = (x11_window_t *)data;

  if (this == NULL)
    return;

  *dest_width  = this->window_class.width;
  *dest_height = this->window_class.height;

  *dest_pixel_aspect = this->display_ratio;
}
  
static void frame_output_cb (void *data,
			     int video_width, int video_height,
			     double video_pixel_aspect,
			     int *dest_x, int *dest_y, 
			     int *dest_width, int *dest_height,
			     double *dest_pixel_aspect,
			     int *win_x, int *win_y) {

  x11_window_t *this = (x11_window_t *)data;
  static int width = 0;
  static int height = 0;
  
  if (this == NULL)
    return;

  if ((width != video_width) || (height != video_height)) {
    oxine_event_t ev;

    width = video_width;
    height = video_height;
#ifdef LOG
    printf("x11: video format changed to %ix%i\n", width, height);
#endif
	    
    ev.type = OXINE_EVENT_FORMAT_CHANGED;
    ev.x = width;
    ev.y = height;
    
    this->event_handler(this->event_handler_data, &ev);
  }

  *dest_x = 0;
  *dest_y = 0;

  *win_x = this->window_class.xpos;
  *win_y = this->window_class.ypos;

  *dest_width  = this->window_class.width;
  *dest_height = this->window_class.height;

  *dest_pixel_aspect = this->display_ratio;
}

/*
 * connect x server, alters following fields in x11_window_t:
 * display, screen, display_ratio, wm_del_win, completion_event
 * returns 1 on success, 0 on error
 */

static int connect_x11_server(x11_window_t *win) {
  
  double    res_h, res_v, aspect_diff;

  if (!XInitThreads ()) {
    printf ("x11: XInitThreads failed\n");
    return 0;
  }

  win->display = XOpenDisplay(NULL);

  if (!win->display) {
    printf("x11: failed open display\n");
    return 0;
  }

  XLockDisplay (win->display);

  /* get screen */
  win->screen = DefaultScreen(win->display);

  /* completion event */
  if (XShmQueryExtension (win->display) == True) {
    win->completion_event = XShmGetEventBase (win->display) + ShmCompletion;
  } else {
    win->completion_event = -1;
  }

  win->display_width = DisplayWidth(win->display, win->screen);
  win->display_height = DisplayHeight(win->display, win->screen);

  /* determine display aspect ratio */ 
  res_h = (DisplayWidth  (win->display, win->screen)*1000
	  / DisplayWidthMM (win->display, win->screen));
  res_v = (DisplayHeight (win->display, win->screen)*1000
	  / DisplayHeightMM (win->display, win->screen));
  win->display_ratio = res_v / res_h;
  aspect_diff = win->display_ratio - 1.0;
  if ((aspect_diff < 0.01) && (aspect_diff > -0.01)) {
    win->display_ratio   = 1.0;
  }

  /* we want to get notified if user closes the window */
  win->wm_del_win = XInternAtom(win->display, "WM_DELETE_WINDOW", False);

  XUnlockDisplay (win->display);

  return 1;
}

/*
 * closes a connection to a x server
 */

static int disconnect_x11_server(x11_window_t *win) {

  if (win->display) {
    XLockDisplay(win->display);
    XCloseDisplay(win->display);
  }

  win->display=NULL;
  win->screen=0;
  win->display_ratio=0;
  win->completion_event=0;

  return 1;
}

/*
 * creates a x window with given geometry, alters
 * window, xpos, ypos, width, height, vis
 * in xs_window_t structure
 * returns 1 on success, 0 on error
 */

static unsigned char bm_no_data[] = { 0,0,0,0, 0,0,0,0 };

static int create_x11_window(x11_window_t *win, int x, int y, int w, int h) {

  int        black;
  Pixmap     bm_no;
  char       window_title[256];
  XSizeHints hint;
  
  snprintf(window_title, 255, "oxine v%s", VERSION);
  
  win->window_class.xpos = x;
  win->window_class.ypos = y;
  win->window_class.width = w;
  win->window_class.height = h;
  
  XLockDisplay (win->display);
  
  black = BlackPixel(win->display, win->screen);
  
  win->window = XCreateSimpleWindow (win->display, 
			    DefaultRootWindow(win->display),
			    win->window_class.xpos, win->window_class.ypos, 
			    win->window_class.width, win->window_class.height, 
			    0, black, black);

  if (!win->window) {
    XUnlockDisplay (win->display);
    return 0;
  }
  
  hint.flags = 0;

  XSetStandardProperties (win->display, win->window, 
			    window_title, window_title, None, NULL, 0, &hint);
  
  XSelectInput (win->display, win->window,
		StructureNotifyMask | 
		ExposureMask |
		ButtonPressMask | 
		PointerMotionMask |
		KeyPressMask );

  XSetWMProtocols(win->display, 
      win->window, 
      &(win->wm_del_win), 
      1);

  bm_no = XCreateBitmapFromData(win->display,
                                win->window,
                                bm_no_data, 8, 8);
  
  win->no_cursor = XCreatePixmapCursor(win->display, bm_no, bm_no,
					(XColor *) &black,
					(XColor *) &black,
					0, 0);
  win->on_cursor = XCreateFontCursor (win->display, XC_left_ptr);
  
  XUnlockDisplay (win->display);

  win->vis.display          = win->display;
  win->vis.screen           = win->screen;
  win->vis.d                = win->window;
  win->vis.dest_size_cb     = dest_size_cb;
  win->vis.frame_output_cb  = frame_output_cb;
  win->vis.user_data        = win;

  return 1;
}

/*
 * deletes a window
 */

static int destroy_x11_window(x11_window_t *win) {

  XLockDisplay (win->display);
  XDestroyWindow(win->display, win->window);
  XUnlockDisplay (win->display);

  win->window_class.xpos = 0;
  win->window_class.ypos = 0;
  win->window_class.width = 0;
  win->window_class.height = 0;
  
  win->vis.display          = NULL;
  win->vis.screen           = 0;
  win->vis.dest_size_cb     = NULL;
  win->vis.frame_output_cb  = NULL;
  win->vis.user_data        = NULL;

  return 1;
}

/*
 * odk job to switch off mouse cursor
 */

static void switch_off(void *data) {

  x11_window_t *win = (x11_window_t*) data;

  if (!win->fullscreen_mode) return;

  if (win->cursor_visible) {
    win->cursor_visible--;
  } else {
    XLockDisplay (win->display);
    XDefineCursor(win->display, win->fullscreen_window,
		  win->no_cursor);
    XFlush(win->display);
    XUnlockDisplay (win->display);
  }
  win->switchoff_job = schedule_job(1000, switch_off, data);
}

/*
 * set fullscreen mode
 */

#define MWM_HINTS_DECORATIONS   (1L << 1)
#define PROP_MWM_HINTS_ELEMENTS 5
typedef struct {
  uint32_t                    flags;
  uint32_t                    functions;
  uint32_t                    decorations;
  int32_t                     input_mode;
  uint32_t                    status;
} MWMHints;

static int fullscreen(odk_window_t *this, int mode) {

  x11_window_t *win = (x11_window_t *)this;

  XSizeHints   hint;
  Atom         prop;
  MWMHints     mwmhints;
  XEvent       xev;
  char         window_title[256];
  
  snprintf(window_title, 255, "oxine v%s - fullscreen", VERSION);

  if (mode == win->fullscreen_mode)
    return 1;

  XLockDisplay (win->display);

  if (mode) {
    
    int black = BlackPixel(win->display, win->screen);

#ifdef LOG
    printf("x11: entering fullscreen mode\n");
#endif

    win->sav_x = win->window_class.xpos;
    win->sav_y = win->window_class.ypos;
    win->sav_w = win->window_class.width;
    win->sav_h = win->window_class.height;
    win->window_class.xpos = 0;
    win->window_class.ypos = 0;
    win->window_class.width = win->display_width;
    win->window_class.height = win->display_height;
    
    win->fullscreen_window = XCreateSimpleWindow (win->display,
						  RootWindow (win->display, win->screen),
						  0, 0,
						  win->display_width, 
						  win->display_height, 1,
						  black, black);

    hint.flags       = PPosition | PSize | PMinSize | PWinGravity;
    hint.x           = 0;
    hint.y           = 0;
    hint.width       = win->display_width;
    hint.height      = win->display_height;
    hint.min_width   = win->display_width;
    hint.min_height  = win->display_height;
    hint.win_gravity = StaticGravity;

    /* set Properties for window manager (always before mapping) */
    XSetStandardProperties (win->display, win->fullscreen_window, 
			    window_title, window_title, None, NULL, 0, &hint);
    
    XSetWMNormalHints (win->display, win->fullscreen_window, &hint);

    /* XSetWMHints (win->display, win->fullscreen_window, &hint);*/

    /*
     * layer above most other things, like gnome panel
     * WIN_LAYER_ABOVE_DOCK  = 10
     *
     */
    {
      static Atom           XA_WIN_LAYER = None;
      long                  propvalue[1];
 
      if (XA_WIN_LAYER == None)
	XA_WIN_LAYER = XInternAtom(win->display, "_WIN_LAYER", False);
      
      propvalue[0] = 20;
      XChangeProperty(win->display, win->fullscreen_window, XA_WIN_LAYER,
		      XA_CARDINAL, 32, PropModeReplace, 
		      (unsigned char *)propvalue, 1);
    }

    /*
     * fullscreen the modern (e.g. metacity) way
     */

    {
      static Atom           XA_WIN_STATE = None;
      long                  propvalue[2];
    
      if (XA_WIN_STATE == None)
	XA_WIN_STATE = XInternAtom (win->display, "_NET_WM_STATE", False);

      propvalue[0] = XInternAtom (win->display, "_NET_WM_STATE_FULLSCREEN", 
				  False);
      propvalue[1] = 0;

      XChangeProperty (win->display, win->fullscreen_window, 
		       XA_WIN_STATE, XA_ATOM, 
		       32, PropModeReplace, 
		       (unsigned char *)propvalue, 1);
      XFlush(win->display);

    }

    /*
     * wm, no borders please
     */

    prop = XInternAtom(win->display, "_MOTIF_WM_HINTS", False);
    mwmhints.flags = MWM_HINTS_DECORATIONS;
    mwmhints.decorations = 0;
    XChangeProperty (win->display, win->fullscreen_window, prop, prop, 32,
		     PropModeReplace, (unsigned char *) &mwmhints,
		     PROP_MWM_HINTS_ELEMENTS);

    XSetTransientForHint(win->display, win->fullscreen_window, None);
    XRaiseWindow(win->display, win->fullscreen_window);

    XSelectInput(win->display, win->fullscreen_window,
		 ExposureMask | KeyPressMask | ButtonPressMask 
		 | StructureNotifyMask | FocusChangeMask | PointerMotionMask);

    XMapWindow (win->display, win->fullscreen_window);

    XFlush(win->display);

    /* Wait for map. */

    do  {
      XMaskEvent(win->display,
		 StructureNotifyMask,
		 &xev) ;
    } while (xev.type != MapNotify || xev.xmap.event != win->fullscreen_window);

    XSetInputFocus (win->display, win->fullscreen_window, 
		    RevertToNone, CurrentTime);

    XMoveWindow (win->display, win->fullscreen_window, 0, 0); 

    xine_gui_send_vo_data (win->window_class.stream,
			   XINE_GUI_SEND_DRAWABLE_CHANGED,
			   (void*)win->fullscreen_window);

    win->cursor_visible = 3;
    XDefineCursor(win->display, win->fullscreen_window,
		      win->on_cursor);
    XFlush(win->display);
    
    win->switchoff_job = schedule_job(1000, switch_off, win);

  } else {

#ifdef LOG
    printf("x11: leaving fullscreen mode\n");
#endif

    xine_gui_send_vo_data (win->window_class.stream,
			   XINE_GUI_SEND_DRAWABLE_CHANGED,
			   (void*)win->window);

    XDestroyWindow (win->display, win->fullscreen_window);

    win->window_class.xpos = win->sav_x;
    win->window_class.ypos = win->sav_y;
    win->window_class.width = win->sav_w;
    win->window_class.height = win->sav_h;

    XFlush(win->display);
  }

  win->fullscreen_mode = mode;

  XUnlockDisplay (win->display);

  return 1;
}

/*
 * public functions
 */

/*
 * x event loop
 */

static void event_loop(odk_window_t *this) {

  x11_window_t *win = (x11_window_t *)this;
  
  win->keep_going    = 1;

  while (win->keep_going) {

    XEvent event;

    XNextEvent (win->display, &event);

    switch (event.type) {
      case Expose:
	{
	  if (event.xexpose.count == 0)
	    xine_gui_send_vo_data (win->window_class.stream,
				 XINE_GUI_SEND_EXPOSE_EVENT, &event);
        }
	break;
      case ConfigureNotify:
	{
	  XConfigureEvent *cev = (XConfigureEvent *) &event;
	  Window tmp_win;

	  if (win->fullscreen_mode) break;

	  win->window_class.width  = cev->width;
	  win->window_class.height = cev->height;

#ifdef LOG
    printf("x11: configure notify w: %i h:%i\n", cev->width, cev->height);
#endif

	  if ((cev->x == 0) && (cev->y == 0)) {
	    XLockDisplay(cev->display);
	    XTranslateCoordinates(cev->display, cev->window, 
				  DefaultRootWindow(cev->display),
				  0, 0, &win->window_class.xpos,
				  &win->window_class.ypos, &tmp_win);

	    XUnlockDisplay(cev->display);
	  }
	  else {
	    win->window_class.xpos = cev->x;
	    win->window_class.ypos = cev->y;
	  }
	}
	break;
	
      case ClientMessage:
	{
	  XClientMessageEvent *cmessage = (XClientMessageEvent *) &event;

	  if ( cmessage->data.l[0] == win->wm_del_win )
	    /* we got a window deletion message from our window manager.*/
	    win->keep_going=0;
	}
	break;
	
      case MotionNotify: 
	{
	  XMotionEvent     *mevent = (XMotionEvent *) &event;
	  x11_rectangle_t   rect;

	  if (win->fullscreen_mode) {
	    lock_job_mutex();
	    win->cursor_visible = 3;
	    XLockDisplay(win->display);
	    XDefineCursor(win->display, win->fullscreen_window,
		      win->on_cursor);
	    XFlush(win->display);
	    XUnlockDisplay(win->display);
	    unlock_job_mutex();
	  }
	  
#ifdef LOG
	  printf("x11: mouse event: mx=%d my=%d\n",mevent->x, mevent->y); 
#endif
	  
	  rect.x = mevent->x;
	  rect.y = mevent->y;
	  rect.w = 0;
	  rect.h = 0;
	  
	  xine_gui_send_vo_data (win->window_class.stream, 
				 XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO, 
				 (void*)&rect);

	  if (win->event_handler) {
	    oxine_event_t ev;
	    
	    ev.type = OXINE_EVENT_MOTION;
	    ev.x = rect.x;
	    ev.y = rect.y;
	    
	    win->event_handler(win->event_handler_data, &ev);
	  }
	
	}
      break;
      
      case ButtonPress: 
	{
	  XButtonEvent *bevent = (XButtonEvent *) &event;
	  x11_rectangle_t   rect;

	  if (win->fullscreen_mode) {
	    lock_job_mutex();
	    win->cursor_visible = 3;
	    XLockDisplay(win->display);
	    XDefineCursor(win->display, win->fullscreen_window,
		      win->on_cursor);
	    XFlush(win->display);
	    XUnlockDisplay(win->display);
	    unlock_job_mutex();
	  }
#ifdef LOG
	  printf("gtkxine: mouse button event: mx=%d my=%d\n",
		 bevent->x, bevent->y); 
#endif
	  rect.x = bevent->x;
	  rect.y = bevent->y;
	  rect.w = 0;
	  rect.h = 0;
	  
	  xine_gui_send_vo_data (win->window_class.stream, 
				 XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO, 
				 (void*)&rect);

	  if (win->event_handler) {
	    oxine_event_t ev;
	    
	    ev.type = OXINE_EVENT_BUTTON;
	    ev.x = rect.x;
	    ev.y = rect.y;
	    
	    switch(bevent->button) {
	      case Button1:
		ev.key = OXINE_BUTTON1;
		break;
	      case Button2:
		ev.key = OXINE_BUTTON2;
		break;
	      case Button3:
		ev.key = OXINE_BUTTON3;
		break;
	      case Button4:
		ev.key = OXINE_BUTTON4;
		break;
	      case Button5:
		ev.key = OXINE_BUTTON5;
		break;
	      default:
		printf("x11: unknown mouse button: %i\n", bevent->button); 
	    }
	    win->event_handler(win->event_handler_data, &ev);
	  }
	}
	break;

      case KeyPress:
	{
	  XKeyEvent      *kevent = (XKeyEvent *) &event;
	  KeySym         key;
	  char           kbuf[256];
	  int            len;

	  XLockDisplay(kevent->display);
	  len = XLookupString(kevent, kbuf, sizeof(kbuf), &key, NULL);
	  XUnlockDisplay(kevent->display);

	  switch (key) {
	    case XK_Escape:
	    case XK_q:
	      win->keep_going = 0;
	      break;
	    case XK_f:
	      if (win->fullscreen_mode)
		fullscreen(this,0);
	      else
		fullscreen(this,1);
	      break;
	  }

	  if (win->event_handler) {
	    oxine_event_t ev;
	    
	    ev.type = OXINE_EVENT_KEY;
	    
	    switch (key) {
	      case XK_F1:
		ev.key = OXINE_KEY_MENU1;
		break;
	      case XK_F2:
		ev.key = OXINE_KEY_MENU2;
		break;
	      case XK_F3:
		ev.key = OXINE_KEY_MENU3;
		break;
	      case XK_F4:
		ev.key = OXINE_KEY_MENU4;
		break;
	      case XK_F5:
		ev.key = OXINE_KEY_MENU5;
		break;
	      case XK_F6:
		ev.key = OXINE_KEY_MENU6;
		break;
	      case XK_F7:
		ev.key = OXINE_KEY_MENU7;
		break;
	      case XK_Next:
		ev.key = OXINE_KEY_NEXT;
		break;
	      case XK_Prior:
		ev.key = OXINE_KEY_PRIOR;
		break;
	      case XK_s:
		ev.key = OXINE_KEY_STOP;
		break;
	      case XK_o:
		ev.key = OXINE_KEY_OSD;
		break;
	      case XK_e:
		ev.key = OXINE_KEY_EJECT;
		break;
	      case XK_m:
		ev.key = OXINE_KEY_VOLMUTE;
		break;
	      case XK_space:
		ev.key = OXINE_KEY_PPLAY;
		break;
	      case XK_p:
		ev.key = OXINE_KEY_PL_PREV;
		break;
	      case XK_n:
		ev.key = OXINE_KEY_PL_NEXT;
		break;
	      case XK_KP_Add:
	      case XK_plus:
		ev.key = OXINE_KEY_VOLUP;
		break;
	      case XK_KP_Subtract:
	      case XK_minus:
		ev.key = OXINE_KEY_VOLDOWN;
		break;
	      case XK_Up:
	      case XK_KP_Up:
		ev.key = OXINE_KEY_UP;
		break;
	      case XK_Down:
	      case XK_KP_Down:
		ev.key = OXINE_KEY_DOWN;
		break;
	      case XK_Left:
	      case XK_KP_Left:
		ev.key = OXINE_KEY_LEFT;
		break;
	      case XK_Right:
	      case XK_KP_Right:
		ev.key = OXINE_KEY_RIGHT;
		break;
	      case XK_Return:
	      case XK_KP_Enter:
		ev.key = OXINE_KEY_SELECT;
		break;
	      case XK_0:
	      case XK_KP_0:
		ev.key = OXINE_KEY_0;
		break;
	      case XK_1:
	      case XK_KP_1:
		ev.key = OXINE_KEY_1;
		break;
	      case XK_2:
	      case XK_KP_2:
		ev.key = OXINE_KEY_2;
		break;
	      case XK_3:
	      case XK_KP_3:
		ev.key = OXINE_KEY_3;
		break;
	      case XK_4:
	      case XK_KP_4:
		ev.key = OXINE_KEY_4;
		break;
	      case XK_5:
	      case XK_KP_5:
		ev.key = OXINE_KEY_5;
		break;
	      case XK_6:
	      case XK_KP_6:
		ev.key = OXINE_KEY_6;
		break;
	      case XK_7:
	      case XK_KP_7:
		ev.key = OXINE_KEY_7;
		break;
	      case XK_8:
	      case XK_KP_8:
		ev.key = OXINE_KEY_8;
		break;
	      case XK_9:
	      case XK_KP_9:
		ev.key = OXINE_KEY_9;
		break;
	      default:
		ev.key = 0;
	    }
	    win->event_handler(win->event_handler_data, &ev);
	  }
	}
	break;

    }
    if ((event.type == win->completion_event) && win->window_class.stream) {
      xine_gui_send_vo_data (win->window_class.stream, 
	  XINE_GUI_SEND_COMPLETION_EVENT, &event);
    }
  }
  win->keep_going=0;
}

static void set_event_handler(odk_window_t *this, odk_cb cb, 
      void *data) {
  
  x11_window_t *win = (x11_window_t *)this;

  win->event_handler = cb;
  win->event_handler_data = data;
}

/*
 * makes a window visible
 * returns 1 on success, 0 on error
 */

static int show(odk_window_t *this) {

  x11_window_t *win = (x11_window_t *)this;

  XLockDisplay (win->display);
  XMapWindow (win->display, win->window);
  XUnlockDisplay (win->display);

  return 1;
}

/*
 * hides a window
 * returns 1 on success, 0 on error
 */

static int hide(odk_window_t *this) {

  x11_window_t *win = (x11_window_t *)this;

  XLockDisplay (win->display);
  XUnmapWindow (win->display, win->window);
  XUnlockDisplay (win->display);

  return 1;
}

/*
 * alters geometry of a window
 * returns 1 on success, 0 on error
 */

static int displace(odk_window_t *this, int x, int y, int w, int h) {

  x11_window_t *win = (x11_window_t *)this;

  if ((this->xpos != x) || (this->ypos != y)) {
    this->xpos = x;
    this->ypos = y;
    XLockDisplay (win->display);
    XMoveWindow (win->display, win->window, x, y);
    XUnlockDisplay (win->display);
  }

  if ((this->width != w) || (this->height != h)) {
    this->width = w;
    this->height = h;
    XLockDisplay (win->display);
    XResizeWindow (win->display, win->window, w, h);
    XUnlockDisplay (win->display);
  }

  return 1;
}


static void stop_event_loop(odk_window_t *this) {

  x11_window_t *win = (x11_window_t *)this;

  XEvent event;

  event.type = Expose;
  event.xexpose.display = win->display;
  event.xexpose.window = win->window;
  event.xexpose.x = 0;
  event.xexpose.y = 0;
  event.xexpose.width = 1;
  event.xexpose.height = 1;

  win->keep_going = 0;

  XLockDisplay (win->display);
  XSendEvent(win->display, win->window, False, ExposureMask, &event);

  XFlush(win->display);
  XUnlockDisplay (win->display);
}


static int dispose(odk_window_t *this) {

  x11_window_t *win = (x11_window_t *)this;

  if (win->keep_going) {

    /* we are in a event loop */
    /* get out of there       */

    printf("x11: dispose: event loop still running!\n");
    stop_event_loop(this);
  }

  /* we need to lock_job_mutex, because our job 
     modifies win->screensaver_job */
  lock_job_mutex();
  cancel_job(win->screensaver_job);
  unlock_job_mutex();

  xine_close_video_driver (this->xine, this->video_port);
  this->video_port = NULL;

  destroy_x11_window(win);
  disconnect_x11_server(win);

  ho_free(win);

  return 1;
}

static void screensaver_job (void *this_gen) {

  x11_window_t *win = (x11_window_t *) this_gen;

  if (win->fullscreen_mode) {
    XLockDisplay (win->display);
#ifdef HAVE_XTESTEXTENSION
    if (win->have_xtest == True) {
      XTestFakeKeyEvent (win->display, win->xtest_keycode, True, CurrentTime);
      XTestFakeKeyEvent (win->display, win->xtest_keycode, False, CurrentTime);
      XSync (win->display, False);
    } else 
#endif
      {
	XResetScreenSaver (win->display);
      }
      XUnlockDisplay (win->display);
  }
  win->screensaver_job = schedule_job(2000, screensaver_job, this_gen);
}

/*
 * constructor
 */

odk_window_t *x11_construct(xine_t *xine, const char *driver) {

  x11_window_t *win = ho_new_tagged(x11_window_t, "x11 window object");
  
  if (!connect_x11_server(win)) {
    ho_free(win);
    return NULL;
  }

  create_x11_window(win, 0, 0, 800, 600);

  win->window_class.xine = xine;

  /* create video port */
  win->window_class.video_port = xine_open_video_driver(xine, 
							driver,
							XINE_VISUAL_TYPE_X11,
							(void *) &(win->vis));
  if (!win->window_class.video_port) {
    printf("x11: failed to open video port '%s', falling back to auto\n", driver);
    win->window_class.video_port = xine_open_video_driver(xine, 
							  NULL,
							  XINE_VISUAL_TYPE_X11,
							  (void *) &(win->vis));
  }

  if (!win->window_class.video_port) {
    printf("x11: failed to open video port\n");
    destroy_x11_window(win);
    ho_free(win);
    return NULL;
  }
  
  win->window_class.show              = show;
  win->window_class.hide              = hide;
  win->window_class.displace          = displace;
  win->window_class.dispose           = dispose;
  win->window_class.fullscreen        = fullscreen;
  win->window_class.event_loop        = event_loop;
  win->window_class.stop_event_loop   = stop_event_loop;
  win->window_class.set_event_handler = set_event_handler;

  /* 
   * screensaver disabling stuff
   */
  {
    int a, b, c, d;

    win->have_xtest = XTestQueryExtension (win->display, &a, &b, &c, &d);
    if (win->have_xtest == True)
      win->xtest_keycode = XKeysymToKeycode (win->display, XK_Shift_L);
  }

  win->screensaver_job = schedule_job(2000, screensaver_job, win);
  
  return (odk_window_t *) win;
}
