cvs_id = "$Id: Config.py,v 1.10 2003/03/16 11:48:19 juri Exp $"

import cPickle
import os
import traceback
from urlparse import urlparse

import straw
from error import log

def ensure_directory(strawdir):
    if os.path.exists(strawdir):
        if not os.path.isdir(strawdir):
            raise Exception
        return 0
    os.mkdir(strawdir)
    return 1

# using gconf seemed a bit too painful for this, seems it's suitable
# for fixed configuration options but not really much else
class ConfigSaver:
    def __init__(self):
        straw.config.signal_connect(
            straw.LocalConfigOptionSignal, self.notify_change)
        self._config_file = os.path.join(straw.config._straw_dir, "config")
        self.load_config()
        return

    def notify_change(self, signal):
        self.save_config()
        return

    def load_config(self):
        if os.path.exists(self._config_file):
            fl = straw.FeedList()
            f = open(self._config_file)
            cf = cPickle.load(f)
            for feed in cf['feeds']:
                nf = straw.Feed("")
                nf.undump(feed)
                fl.append(nf)
            straw.config.feeds = fl
            straw.config.poll_frequency = cf.get('poll_frequency', 1200)
            straw.config.number_of_items_stored = cf.get(
                'number_of_items_stored', 30)
            straw.config.item_order = cf.get('item_order', 1)
            straw.config.feed_id_seq = cf.get('feed_id_seq', 0)
            straw.config.main_window_size = cf.get('main_window_size', (700, 550))
            straw.config.offline = cf.get('offline', 0)
            # config.poll_spread 
            f.close()
        return

    def save_config(self):
        cf = {}
        l = []
        for f in straw.config.feeds:
            l.append(f.dump())
        cf['feeds'] = l
        cf['poll_frequency'] = straw.config.poll_frequency
        cf['number_of_items_stored'] = straw.config.number_of_items_stored
        cf['item_order'] = straw.config.item_order
        cf['feed_id_seq'] = straw.config.feed_id_seq
        cf['main_window_size'] = straw.config.main_window_size
        cf['offline'] = straw.config.offline
        pickle_file = open(self._config_file, 'w')
        cPickle.dump(cf, pickle_file, 1)
        pickle_file.close()
        return

class BaseProxyConfig(object):
    def get_config_working(self):
        return self._config_working

    config_working = property(get_config_working, None, None, "")

    def get_host(self):
        return self._host

    def set_host(self, host):
        self._host = host
        self._ip = None
        if self._host:
            straw.lookup_manager.lookup(self._host, self.ip_lookup_callback)

    host = property(get_host, set_host, None, "")

    def get_ip(self):
        return self._ip

    def set_ip(self, ip):
        assert self._ip is None, "Overwriting proxy IP not allowed!"
        if ip is None:
            self.use = 0
        self._ip = ip

    ip = property(get_ip, set_ip, None, "")

    def ip_lookup_callback(self, host, ip, data):
        self.ip = ip

    def _is_waiting(self):
        return self.use and not self._ip

    is_waiting = property(_is_waiting, None, None, "")

class GconfProxyConfig(BaseProxyConfig):
    """Encapsulate proxy use and location information (host, port, ip),
    gconf reading and name lookup logic"""

    GCONF_PROXY_ROOT = "/system/proxy"
    GCONF_PROXY_MODE = GCONF_PROXY_ROOT + "/mode"
    GCONF_HTTP_PROXY_ROOT = "/system/http_proxy"
    GCONF_HTTP_PROXY_HOST = GCONF_HTTP_PROXY_ROOT + "/host"
    GCONF_HTTP_PROXY_PORT = GCONF_HTTP_PROXY_ROOT + "/port"
    GCONF_HTTP_PROXY_USE_AUTHENTICATION = GCONF_HTTP_PROXY_ROOT + "/use_authentication"
    GCONF_HTTP_PROXY_USER = GCONF_HTTP_PROXY_ROOT + "/authentication_user"
    GCONF_HTTP_PROXY_PASSWORD = GCONF_HTTP_PROXY_ROOT + "/authentication_password"

    def __init__(self):
        try:
            import gconf
        except ImportError:
            log(_("No gconf module installed, not reading proxy configuration from there"))
            self._config_working = 0
            return
        client = gconf.client_get_default()
        self._config_working = client.dir_exists(self.GCONF_PROXY_ROOT) and client.dir_exists(self.GCONF_HTTP_PROXY_ROOT)
        if not self._config_working:
            return
        client.add_dir(self.GCONF_PROXY_ROOT, gconf.CLIENT_PRELOAD_RECURSIVE)
        client.add_dir(self.GCONF_HTTP_PROXY_ROOT, gconf.CLIENT_PRELOAD_RECURSIVE)
        client.notify_add(self.GCONF_PROXY_MODE, self.proxy_mode_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_HOST, self.proxy_host_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_PORT, self.proxy_port_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_USE_AUTHENTICATION,
                          self.proxy_auth_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_USER, self.proxy_auth_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_PASSWORD, self.proxy_auth_changed)

        self._ip = None
        pm = client.get_string(self.GCONF_PROXY_MODE)
        self.use = pm is not None and pm != "none"
        self.host = client.get_string(self.GCONF_HTTP_PROXY_HOST)
        self.port = client.get_int(self.GCONF_HTTP_PROXY_PORT)
        self.use_authentication = client.get_bool(
            self.GCONF_HTTP_PROXY_USE_AUTHENTICATION)
        self.user = client.get_string(self.GCONF_HTTP_PROXY_USER)
        self.password = client.get_string(self.GCONF_HTTP_PROXY_PASSWORD)

    # here be gconf logic
    def proxy_mode_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_string()
        self.use = value is not None and value != "none"

    def proxy_host_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_string()
        if value is not None and len(value):
            self.host = value
        else:
            self.use = 0

    def proxy_port_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_int()
        if value is not None and value > 0:
            self.port = value
        else:
            self.use = 0

    def proxy_auth_changed(self, client, notify_id, entry, *args):
        value = entry.value
        if entry.key == self.GCONF_HTTP_PROXY_USE_AUTHENTICATION:
            if not value:
                self.use_authentication = 0
            else:
                self.use_authentication = 1
        elif entry.key == self.GCONF_HTTP_PROXY_USER:
            self.user = value
        elif entry.key == self.GCONF_HTTP_PROXY_PASSWORD:
            self.password = value

class EnvironmentProxyConfig(BaseProxyConfig):
    """Encapsulate proxy use and location information, environment reading"""

    def __init__(self):
        try:
            result = self._read_env()
        except:
            result = 0
        self._config_working = result
        self.use = self._config_working

    def _read_env(self):
        proxy = os.getenv("http_proxy")
        if not proxy:
            return 0
        pparts = urlparse(proxy)
        proto = pparts[0]
        host = pparts[1]
        if proto != "http" or not len(host):
            return 0
        hparts = host.split('@')
        auth = self.user = self.password = self.use_authentication = None
        if len(hparts) > 2:
            return 0
        if len(hparts) == 2:
            auth, host = hparts
        if auth:
            aparts = auth.split(':')
            if len(aparts) != 2:
                return 0
            self.user, self.password = aparts
            self.use_authentication = 1
        self.host = host
        self.port = 80
        hparts = host.split(':')
        if len(hparts) > 2:
            return 0
        if len(hparts) == 2:
            self.host = hparts[0]
            self.port = int(hparts[1])
        return 1

class NullProxyConfig(BaseProxyConfig):
    def __init__(self):
        self.use = 0
        self._config_working = 0
        self._host = None
        self._ip = None
        self.port = None
        self.user = None
        self.password = None
        self.use_authentication = None

class Config(object, straw.SignalEmitter):
    def __init__(self):
        straw.SignalEmitter.__init__(self)
        self.initialize_slots(straw.FeedsChangedSignal,
                              straw.PollFrequencyChangedSignal,
                              straw.NumberOfItemsStoredChangedSignal,
                              straw.ItemOrderChangedSignal,
                              straw.OfflineModeChangedSignal)
        self._poll_frequency = 1200
        self._poll_spread = 0
        self._feed_id_seq = 0
        self._feeds = straw.FeedList()
        self._feeds.signal_connect(straw.FeedsChangedSignal,
                                   self.feeds_changed)
        self._number_of_items_stored = 30
        self._item_order = 1
        self._straw_dir = os.path.join(os.getenv("HOME"), ".straw")
        self._main_window_size = (700, 550)
        self._offline = 0
        pcchoices = (GconfProxyConfig, EnvironmentProxyConfig, NullProxyConfig)
        for p in pcchoices:
            self._proxy_config = p()
            if self._proxy_config.config_working:
                break
        self.first_time = ensure_directory(self._straw_dir)
        return

    def get_proxy_config(self):
        return self._proxy_config

    proxy_config = property(get_proxy_config, None, None, "")

    # and the rest of the stuff
    def feeds_changed(self, signal):
        self.emit_signal(straw.FeedsChangedSignal(self))
    
    def set_feeds(self, feeds):
        self._feeds = straw.FeedList(feeds)
        self._feeds.signal_connect(straw.FeedsChangedSignal,
                                   self.feeds_changed)
        self.emit_signal(straw.FeedsChangedSignal(self))
        return

    def get_feeds(self):
        return self._feeds

    feeds = property(get_feeds, set_feeds, None, "Feeds")

    def set_poll_frequency(self, poll_frequency):
        self._poll_frequency = poll_frequency
        self.emit_signal(straw.PollFrequencyChangedSignal(self))
        return

    def get_poll_frequency(self):
        return self._poll_frequency

    poll_frequency = property(get_poll_frequency, set_poll_frequency, None,
                              "Polling frequency")

    def get_number_of_items_stored(self):
        return self._number_of_items_stored

    def set_number_of_items_stored(self, num):
        self._number_of_items_stored = num
        self.emit_signal(straw.NumberOfItemsStoredChangedSignal(self))

    number_of_items_stored = property(get_number_of_items_stored,
                                      set_number_of_items_stored, None,
                                      "Number of items to store per feed")

    def get_item_order(self):
        return self._item_order

    def set_item_order(self, order):
        self._item_order = order
        self.emit_signal(straw.ItemOrderChangedSignal(self))

    item_order = property(get_item_order, set_item_order, "Ordering of items")

    def get_feed_id_seq(self):
        self._feed_id_seq += 1
        return self._feed_id_seq

    def set_feed_id_seq(self, id):
        self._feed_id_seq = id
        return

    feed_id_seq = property(get_feed_id_seq, set_feed_id_seq, None, "")

    def set_main_window_size(self, size):
        if size != self._main_window_size:
            self._main_window_size = size
            self.emit_signal(straw.MainWindowSizeChangedSignal(self))

    def get_main_window_size(self):
        return self._main_window_size

    main_window_size = property(get_main_window_size, set_main_window_size, "")

    def get_offline(self):
        return self._offline

    def set_offline(self, mode):
        if self._offline != mode:
            self._offline = mode
            self.emit_signal(straw.OfflineModeChangedSignal(self))

    offline = property(get_offline, set_offline, None, "")
