# Feed wrapper module
# Copyright (c) 2002 Juri Pakaste
# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

cvs_id = "$Id: Feed.py,v 1.20 2003/03/09 16:41:52 juri Exp $"

import urlparse
from threading import RLock
try:
    from cStringIO import StringIO
except:
    import StringIO
import straw
from error import log

class Feed(object, straw.SignalEmitter):
    "A Feed object stores information set by user about a RSS feed."
    __slots__ = ('_title', '_location', '_username', '_password', '_parsed',
                 '__save_fields', '_items', '_idseq', '_idseqlock', '_slots',
                 '_reverse', '_id', 'invalid_rss', '_channel_description',
                 '_channel_title', '_channel_link', '_channel_copyright',
                 'channel_lbd', 'channel_editor', 'channel_webmaster',
                 '_channel_subscriptions', 'channel_subscriptions_location',
                 '_channel_blogroll', 'channel_blogroll_location',
                 'channel_blink', '_previous_etag', '_error')

    __save_fields = (('_title', ""), ('_location', ""), ('_username', ""),
                     ('_password', ""), ('_id', ""),
                     ('_channel_description', ""), ('_channel_title', ""),
                     ('_channel_link', ""), ('_channel_copyright', ""),
                     ('_previous_etag', ""), ('_error', None))
    
    def __init__(self, title, location = "", username = "", password = ""):
        straw.SignalEmitter.__init__(self)
        self.initialize_slots(straw.NewItemsSignal,
                              straw.ItemsAddedSignal,
                              straw.ItemReadSignal,
                              straw.ItemStickySignal,
                              straw.ItemDeletedSignal,
                              straw.AllItemsReadSignal,
                              straw.RefreshFeedDisplaySignal,
                              straw.FeedsChangedSignal,
                              straw.FeedPolledSignal)
        self._id = straw.config.feed_id_seq
        self._title = title
        self._channel_description = ""
        self._channel_title = ""
        self._channel_link = ""
        self._channel_copyright = ""
        self.channel_lbd = None
        self.channel_editor = ""
        self.channel_webmaster = ""
        self._channel_subscriptions = ""
        self._channel_blogroll = ""
        self.channel_subscriptions_location = ""
        self.channel_blogroll_location = ""
        self.channel_blink = ""
        self._location = location
        self._username = username
        self._password = password
        self._parsed = None
        self._items = straw.FeedQD()
        self._items.set_reverse(straw.config.item_order)
        self._reverse = straw.config.item_order
        self._idseq = 0
        self._idseqlock = RLock()
        self.invalid_rss = -1
        straw.config.signal_connect(straw.ItemOrderChangedSignal,
                                    self.item_order_changed)
        self._previous_etag = None
        self._error = None
        return

    def item_order_changed(self, event):
        self._items.set_reverse(event.sender.item_order)
        self._reverse = event.sender.item_order
        self.emit_signal(straw.RefreshFeedDisplaySignal(self))

    def set_parsed(self, parsed):
        self._parsed = parsed
        return

    def get_parsed(self):
        return self._parsed

    # ooh, properties are crunchy
    parsed = property(get_parsed, set_parsed, None,
                      "A ParsedSummary object generated from the summary file")

    def get_id(self):
        return self._id

    id = property(get_id, None, None, "")
    
    def get_title(self):
        return self._title

    def set_title(self, title):
        self._title = title
        self.emit_signal(straw.FeedsChangedSignal(self))
        return

    title = property(get_title, set_title, None,
                     "The title of this Feed (as defined by user)")

    def set_access_info(self, (location, username, password)):
        self._location = location
        self._username = username
        self._password = password
        self.emit_signal(straw.FeedsChangedSignal(self))
        return

    def get_access_info(self):
        return (self._location, self._username, self._password)

    access_info = property(get_access_info, set_access_info, None,
                           "A tuple of location, username, password")

    def get_channel_title(self):
        return self._channel_title

    def set_channel_title(self, t):
        changed = self._channel_title != t
        self._channel_title = t
        if changed:
            self.emit_signal(straw.FeedsChangedSignal(self))

    channel_title = property(get_channel_title, set_channel_title, None, "")

    def get_channel_description(self):
        return self._channel_description

    def set_channel_description(self, t):
        changed = self._channel_description != t
        self._channel_description = t
        if changed:
            self.emit_signal(straw.FeedsChangedSignal(self))

    channel_description = property(get_channel_description, set_channel_description, None, "")

    def get_channel_link(self):
        return self._channel_link

    def set_channel_link(self, t):
        changed = self._channel_link != t
        self._channel_link = t
        if changed:
            self.emit_signal(straw.FeedsChangedSignal(self))

    channel_link = property(get_channel_link, set_channel_link, None, "")

    def get_channel_copyright(self):
        return self._channel_copyright

    def set_channel_copyright(self, t):
        changed = self._channel_copyright != t
        self._channel_copyright = t
        if changed:
            self.emit_signal(straw.FeedsChangedSignal(self))

    channel_copyright = property(get_channel_copyright, set_channel_copyright, None, "")

    def dump(self):
        fl = {}
        for f, default in self.__save_fields:
            fl[f] = self.__getattribute__(f)
        return fl

    def undump(self, dict):
        for f, default in self.__save_fields:
            self.__setattr__(f, dict.get(f, default))
        return

    def complete_url(self, url):
        if urlparse.urlparse(url)[0] == '':
            return urlparse.urljoin(self._location, url)
        else:
            return url

    def next_id(self):
        self._idseqlock.acquire()
        self._idseq += 1
        id = self._idseq
        self._idseqlock.release()
        return id

    def poll_done(self):
        self.emit_signal(straw.FeedPolledSignal(self))

    def add_items(self, new_items):
        really_new, deleted = self._add_items(new_items)
        for d in deleted:
            if d not in really_new:
                self.emit_signal(straw.ItemDeletedSignal(self, d))
        really_new = [ r for r in really_new if r not in deleted ]
        if len(really_new) > 0:
            self.emit_signal(straw.NewItemsSignal(self, really_new))
        
    def restore_items(self, new_items):
        accepted, deleted = self._add_items(new_items)
        accepted = [a for a in accepted if a not in deleted]
        ids = [i.id for i in new_items]
        if len(ids) > 0:
            self._idseq = max(ids)
        self.emit_signal(straw.ItemsAddedSignal(self, new_items))

    def _add_items(self, items):
        old = 0
        accepted = []
        for nitem in items:
            for id, item in self._items.items():
                if item.title == nitem.title and item.description == nitem.description:
                    old = 1
                    self._items.set_item(id, item)
                    break
                elif item.guid is not None and item.guid == nitem.guid:
                    nitem.id = item.id
                    self._items.set_item(id, nitem)
                    nitem.signal_connect(straw.ItemReadSignal,
                                         self.forward_signal)
                    nitem.signal_connect(straw.ItemStickySignal,
                                         self.forward_signal)
                    nitem.feed = self
                    accepted.append(nitem)
                    old = 1
                    break
            if not old:
                if nitem.id is None:
                    nitem.id = self.next_id()
                self._items.set_item(nitem.id, nitem)
                nitem.signal_connect(straw.ItemReadSignal, self.forward_signal)
                nitem.signal_connect(straw.ItemStickySignal, self.forward_signal)
                nitem.feed = self
                accepted.append(nitem)
            old = 0
        removed = self._items.cut(straw.config.number_of_items_stored)
        return (accepted, removed)

    def forward_signal(self, signal):
        self.emit_signal(signal)

    def get_items(self):
        ilist = []
        for item in self._items:
            ilist.append(item)
        return ilist

    def get_next_item(self, item):
        if self._reverse:
            id = self._items.get_next_smaller_key(item.id)
        else:
            id = self._items.get_next_larger_key(item.id)
        return self._items.get(id, None)

    def get_previous_item(self, item):
        if self._reverse:
            id = self._items.get_next_larger_key(item.id)
        else:
            id = self._items.get_next_smaller_key(item.id)
        return self._items.get(id, None)

    def get_first_item(self):
        return self._items.get_first_item()

    def get_last_item(self):
        return self._items.get_last_item()

    def get_item(self, id):
        return self._items[id]

    def get_item_index(self, item):
        idx = self._items.index(item.id)
        if self._reverse:
            idx = len(self._items) - idx -1
            return idx
        return idx

    items = property(get_items, None, None, "Stored items, newest first")

    def number_of_unread(self):
        count = 0
        for i in self._items.values():
            if not i.seen:
                count += 1
        return count

    def __str__(self):
        return "Feed '%s' from %s" % (self._title, self._location)

    def mark_all_read(self):
        changed = []
        i = 0
        for item in self._items:
            if item.set_seen_quiet():
                changed.append((i, item))
            i += 1
        self.emit_signal(straw.AllItemsReadSignal(self, changed))
        return

    def get_previous_etag(self):
        return self._previous_etag

    def set_previous_etag(self, etag):
        self._previous_etag = etag

    previous_etag = property(get_previous_etag, set_previous_etag, None, "")

    def set_error(self, error):
        self._error = error
        self.emit_signal(straw.FeedsChangedSignal(self))

    def get_error(self):
        return self._error

    error = property(get_error, set_error, None, "")

    def get_channel_blogroll(self):
        return self._channel_blogroll

    def set_channel_blogroll(self, data):
        if data and len(data):
            self._channel_blogroll = straw.OPMLImport.read(StringIO(data))

    channel_blogroll = property(get_channel_blogroll, set_channel_blogroll,
                                None, "")

    def get_channel_subscriptions(self):
        return self._channel_subscriptions

    def set_channel_subscriptions(self, data):
        if data and len(data):
            self._channel_subscriptions = straw.OPMLImport.read(StringIO(data))

    channel_subscriptions = property(get_channel_subscriptions,
                                     set_channel_subscriptions, None, "")
